#!/usr/bin/python

"""
_DBCreator_

Base class for formatters that create tables.

"""

from builtins import str

from WMCore.Database.DBFormatter import DBFormatter
from WMCore.WMException import WMException
from WMCore.WMExceptions import WMEXCEPTION


class DBCreator(DBFormatter):

    """
    _DBCreator_

    Generic class for creating database tables.

    """

    def __init__(self, logger, dbinterface):
        """
        _init_

        Call the constructor of the parent class and create empty dictionaries
        to hold table create statements, constraint statements and insert
        statements.
        """
        DBFormatter.__init__(self, logger, dbinterface)
        self.create = {}
        self.constraints = {}
        self.inserts = {}
        self.indexes = {}

    def execute(self, conn = None, transaction = False):
        """
        _execute_

        Generic method to create tables and constraints by execute
        sql statements in the create, and constraints dictionaries.

        Before execution the keys assigned to the tables in the self.create
        dictionary are sorted, to offer the possibilitiy of executing
        table creation in a certain order.

        """
        # create tables
        for i in sorted(self.create.keys()):
            try:
                self.dbi.processData(self.create[i],
                                     conn = conn,
                                     transaction = transaction)
            except Exception as e:
                msg = WMEXCEPTION['WMCORE-2'] + '\n\n' +\
                                  str(self.create[i]) +'\n\n' +str(e)
                self.logger.debug( msg )
                raise WMException(msg,'WMCORE-2')

        # create indexes
        for i in self.indexes:
            try:
                self.dbi.processData(self.indexes[i],
                                     conn = conn,
                                     transaction = transaction)
            except Exception as e:
                msg = WMEXCEPTION['WMCORE-2'] + '\n\n' +\
                                  str(self.indexes[i]) +'\n\n' +str(e)
                self.logger.debug( msg )
                raise WMException(msg,'WMCORE-2')

        # set constraints
        for i in self.constraints:
            try:
                self.dbi.processData(self.constraints[i],
                                 conn = conn,
                                 transaction = transaction)
            except Exception as e:
                msg = WMEXCEPTION['WMCORE-2'] + '\n\n' +\
                                  str(self.constraints[i]) +'\n\n' +str(e)
                self.logger.debug( msg )
                raise WMException(msg,'WMCORE-2')

        # insert permanent data
        for i in self.inserts:
            try:
                self.dbi.processData(self.inserts[i],
                                     conn = conn,
                                     transaction = transaction)
            except Exception as e:
                msg = WMEXCEPTION['WMCORE-2'] + '\n\n' +\
                                  str(self.inserts[i]) +'\n\n' +str(e)
                self.logger.debug( msg )
                raise WMException(msg,'WMCORE-2')

        return True

    def __str__(self):
        """
        _str_

        Return a well formatted text representation of the schema held in the
        self.create, self.constraints, self.inserts, self.indexes dictionaries.
        """
        string = ''
        for i in self.create, self.constraints, self.inserts, self.indexes:
            for j in i:
                string = string + i[j].lstrip() + '\n'
        return string



#!/usr/bin/env python

"""
URLFetcher

A unittest for seeing if we can pull code from a URL using the URLFetcher
Written by someone who has no idea what URLFetcher is supposed to do.
"""
import os
import re
import os.path
import unittest

from WMCore.WMSpec import WMTask
from WMCore.WMSpec import WMStep
from WMCore.WMSpec.Steps.Fetchers.URLFetcher import URLFetcher
from WMQuality.TestInitCouchApp              import TestInitCouchApp as TestInit

class URLFetcherTest(unittest.TestCase):
    """
    Main test for the URLFetcher

    """

    def setUp(self):
        """
        Basic setUp

        """
        self.testInit = TestInit(__file__)
        self.testInit.setLogging()
        self.testInit.setDatabaseConnection()

        self.testDir = self.testInit.generateWorkDir()

        return

    def tearDown(self):
        """
        Basic tearDown

        """
        self.testInit.delWorkDir()

        return


    def createTask(self, fileURL):
        """
        _createTask_

        Create a test task that includes the
        fileURL
        """

        task = WMTask.makeWMTask("testTask")
        task.makeStep("step1")
        task.makeStep("step2")

        for t in task.steps().nodeIterator():
            t = WMStep.WMStepHelper(t)
            os.mkdir('%s/%s' % (self.testDir, t.name()))
            t.data.sandbox.section_('file0')
            t.data.sandbox.file0.src = fileURL
        return task


    def testA_BasicFunction(self):
        """
        _BasicFunction_

        Test and see if we can retrieve a basic file

        URL should not have http:// prefix in it
        """
        url     = 'cmsweb.cern.ch'
        task    = self.createTask(fileURL = 'http://%s' % url)
        fetcher = URLFetcher()
        fetcher.setWorkingDirectory(workingDir = self.testDir)
        fetcher(wmTask = task)

        f = open(os.path.join(self.testDir, 'step2', url))
        content = f.read()
        f.close()

        for x in ['html', 'CMS']:
            self.assertNotEqual( content.find(x), -1 )


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
"""
_DBCore_t_

Unit tests for the DBInterface class
"""




from builtins import range

import unittest
import threading

from WMQuality.TestInit import TestInit

class DBCoreTest(unittest.TestCase):
    def setUp(self):
        self.testInit = TestInit(__file__)
        self.testInit.setLogging()
        self.testInit.setDatabaseConnection()
        self.testInit.setSchema(customModules = ["WMQuality.TestDB"],
                                useDefault = False)

        return

    def tearDown(self):
        """
        Delete the databases
        """
        self.testInit.clearDatabase()
        return

    def testBuildBinds(self):
        """
        Test class for DBCore.buildbinds()

        """

        #This class may become obselete soon.  There is a TODO stuck onto DBCore.buildbinds()
        #This just checks to see that the sequence properly packages the first value is set
        #So that it sets the key seqname to the proper name in the files list
        #Also right now tests that it sets the keys right in each dict, but this seems redundant
        # -mnorman

        seqname       = 'file'
        dictbinds     = {'lumi':123, 'test':100}
        files         = ['testA', 'testB', 'testC']

        myThread = threading.currentThread()

        testInterface = myThread.dbi

        binds         = testInterface.buildbinds(files, seqname, dictbinds)

        #This should return a dict for every value in files with additional elements
        #from dictbinds.  We then loop over every dict (which should be equal to the
        #number of elements in files), and look to see that the filename matches and that
        #At least one of the dictbinds keys matches to its proper element

        for i in range(len(files)):
            self.assertEqual(binds[i][seqname], files[i])
            self.assertEqual(binds[i][next(iter(dictbinds))], dictbinds[next(iter(dictbinds))])


        return

    def testProcessDataNoBinds(self):
        """
        _testProcessDataNoBinds_

        Verify that insert and select queries work when no binds are used.
        """
        insertSQL = "INSERT INTO test_tablea VALUES (1, 2, 'three')"
        selectSQL = "SELECT column1, column2, column3 from test_tablea"

        myThread = threading.currentThread()
        myThread.dbi.processData(insertSQL)
        resultSets = myThread.dbi.processData(selectSQL)

        assert len(resultSets) == 1, \
               "Error: Wrong number of ResultSets returned."

        results = resultSets[0].fetchall()

        assert len(results) == 1, \
               "Error: Wrong number of rows returned."
        assert len(results[0]) == 3, \
               "Error: Wrong number of columns returned."
        assert results[0][0] == 1, \
               "Error: Column one is wrong."
        assert results[0][1] == 2, \
               "Error: Column two is wrong."
        assert results[0][2] == "three", \
               "Error: Column three is wrong."

        return

    def testProcessDataOneBind(self):
        """
        _testProcessDataOneBind_

        Verify that insert and select queries work with one set of bind variables.
        """
        bindsA = {"one": 1, "two": 2, "three": "three"}
        bindsB = {"one": 3, "two": 2, "three": "one"}
        insertSQL = "INSERT INTO test_tablea VALUES (:one, :two, :three)"
        selectSQL = \
          """SELECT column1, column2, column3 FROM test_tablea
             WHERE column1 = :one AND column2 = :two AND column3 = :three"""

        myThread = threading.currentThread()
        myThread.dbi.processData(insertSQL, binds = bindsA)
        myThread.dbi.processData(insertSQL, binds = bindsB)

        resultSets = myThread.dbi.processData(selectSQL, bindsA)

        assert len(resultSets) == 1, \
               "Error: Wrong number of ResultSets returned."

        results = resultSets[0].fetchall()

        assert len(results) == 1, \
               "Error: Wrong number of rows returned."
        assert len(results[0]) == 3, \
               "Error: Wrong number of columns returned."
        assert results[0][0] == 1, \
               "Error: Column one is wrong."
        assert results[0][1] == 2, \
               "Error: Column two is wrong."
        assert results[0][2] == "three", \
               "Error: Column three is wrong."

        resultSets = myThread.dbi.processData(selectSQL, bindsB)

        assert len(resultSets) == 1, \
               "Error: Wrong number of ResultSets returned."

        results = resultSets[0].fetchall()

        assert len(results) == 1, \
               "Error: Wrong number of rows returned."
        assert len(results[0]) == 3, \
               "Error: Wrong number of columns returned."
        assert results[0][0] == 3, \
               "Error: Column one is wrong."
        assert results[0][1] == 2, \
               "Error: Column two is wrong."
        assert results[0][2] == "one", \
               "Error: Column three is wrong."

        return

    def testProcessDataSeveralBinds(self):
        """
        _testProcessDataSeveralBinds_

        Verify that insert and select queries work with several binds.
        """
        bindsA = [{"one": 1, "two": 2, "three": "three"},
                  {"one": 3, "two": 2, "three": "one"},
                  {"one": 4, "two": 5, "three": "six"},
                  {"one": 6, "two": 5, "three": "four"}]
        bindsB = [{"one": 10, "two": 11, "three": "twelve"},
                  {"one": 12, "two": 11, "three": "ten"}]

        insertSQL = "INSERT INTO test_tablea VALUES (:one, :two, :three)"
        selectSQL = \
          """SELECT column1, column2, column3 FROM test_tablea
             WHERE column1 = :one AND column2 = :two AND column3 = :three"""

        myThread = threading.currentThread()
        myThread.dbi.processData(insertSQL, binds = bindsA)
        myThread.dbi.processData(insertSQL, binds = bindsB)

        resultSets = myThread.dbi.processData(selectSQL, bindsA)

        assert len(resultSets) == 1, \
               "Error: Wrong number of ResultSets returned."

        results = resultSets[0].fetchall()

        assert len(results) == 4, \
               "Error: Wrong number of rows returned."

        for result in results:
            assert len(result) == 3, \
                   "Error: Wrong number of columns returned."
            for bind in bindsA:
                if bind["one"] == result[0] and bind["two"] == result[1] and \
                   bind["three"] == result[2]:
                    bindsA.remove(bind)
                    break

        assert len(bindsA) == 0, \
               "Error: Missing rows from select."

        resultSets = myThread.dbi.processData(selectSQL, bindsB)

        assert len(resultSets) == 1, \
               "Error: Wrong number of ResultSets returned."

        results = resultSets[0].fetchall()

        assert len(results) == 2, \
               "Error: Wrong number of rows returned."

        for result in results:
            assert len(result) == 3, \
                   "Error: Wrong number of columns returned."
            for bind in bindsB:
                if bind["one"] == result[0] and bind["two"] == result[1] and \
                   bind["three"] == result[2]:
                    bindsB.remove(bind)
                    break

        assert len(bindsB) == 0, \
               "Error: Missing rows from select."

        return

    def testProcessDataHugeBinds(self):
        """
        _testProcessDataHugeBinds_

        Verify that select and insert queries work with several thousand binds.
        """
        bindsA = []
        bindsB = []
        for i in range(3001):
            bindsA.append({"one": i, "two": i * 2, "three": str(i * 3)})

        for i in range(1501):
            bindsB.append({"one": (i + 1) * 2, "two": i, "three": str(i * 5)})

        insertSQL = "INSERT INTO test_tablea VALUES (:one, :two, :three)"
        selectSQL = \
          """SELECT column1, column2, column3 FROM test_tablea
             WHERE column1 = :one AND column2 = :two AND column3 = :three"""

        myThread = threading.currentThread()
        myThread.dbi.processData(insertSQL, binds = bindsA)
        myThread.dbi.processData(insertSQL, binds = bindsB)

        resultSets = myThread.dbi.processData(selectSQL, bindsA)
        results = []
        for resultSet in resultSets:
            results.extend(resultSet.fetchall())

        assert len(results) == 3001, \
               "Error: Wrong number of rows returned: %d" % len(results)

        for result in results:
            assert len(result) == 3, \
                   "Error: Wrong number of columns returned."
            for bind in bindsA:
                if bind["one"] == result[0] and bind["two"] == result[1] and \
                   bind["three"] == result[2]:
                    bindsA.remove(bind)
                    break

        assert len(bindsA) == 0, \
               "Error: Missing rows from select."

        resultSets = myThread.dbi.processData(selectSQL, bindsB)
        results = []
        for resultSet in resultSets:
            results.extend(resultSet.fetchall())

        assert len(results) == 1501, \
               "Error: Wrong number of rows returned."

        for result in results:
            assert len(result) == 3, \
                   "Error: Wrong number of columns returned."
            for bind in bindsB:
                if bind["one"] == result[0] and bind["two"] == result[1] and \
                   bind["three"] == result[2]:
                    bindsB.remove(bind)
                    break

        assert len(bindsB) == 0, \
               "Error: Missing rows from select."

        return

    def testInsertHugeNumber(self):
        """
        _testInsertHugeNumber_

        Verify that we can insert and select huge numbers.
        """
        insertSQL = "INSERT INTO test_bigcol VALUES(:val1)"
        selectSQL = "SELECT * FROM test_bigcol"

        bindsA = {"val1": 2012211901}
        bindsB = {"val1": 20122119010}

        myThread = threading.currentThread()
        myThread.dbi.processData(insertSQL, binds = bindsA)
        myThread.dbi.processData(insertSQL, binds = bindsB)

        resultSets = myThread.dbi.processData(selectSQL)
        results = []
        for resultSet in resultSets:
            for row in resultSet.fetchall():
                results.append(row[0])

        assert len(results) == 2, \
               "Error: Wrong number of results."
        assert bindsA["val1"] in results, \
               "Error: Value one is missing."
        assert bindsB["val1"] in results, \
               "Error: Value one is missing."

        return

if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python

import random

from WMCore.FwkJobReport import Report
from WMCore.DataStructs.Run import Run
from WMCore.Services.UUIDLib import makeUUID

outputModules = ["write_A_Calo_RAW", "write_A_Cosmics_RAW",
                 "write_A_HcalHPDNoise_RAW", "write_A_MinimumBias_RAW",
                 "write_A_RandomTriggers_RAW", "write_A_Calibration_TestEnables_RAW",
                 "write_HLTDEBUG_Monitor_RAW"]

runInfo = Run(1)
runInfo.extendLumis([11, 12, 13, 14, 15])

for i in range(100):
    loadTestReport = Report.Report("cmsRun1")
    loadTestReport.addInputSource("PoolSource")
    inputFile = loadTestReport.addInputFile("PoolSource", lfn = makeUUID(),
                                            events = 600000, size = 600000)
    Report.addRunInfoToFile(inputFile, runInfo)

    for outputModule in outputModules:
        loadTestReport.addOutputModule(outputModule)
        datasetInfo = {"applicationName": "cmsRun", "applicationVersion": "CMSSW_3_3_5_patch3",
                       "primaryDataset": outputModule, "dataTier": "RAW",
                       "processedDataset": "LoadTest10"}
        fileAttrs = {"lfn": makeUUID(), "location": "cmssrm.fnal.gov",
                     "checksums": {"adler32": "ff810ec3", "cksum": "2212831827"},
                     "events": random.randrange(500, 5000, 50),
                     "merged": random.choice([True, False]),
                     "size": random.randrange(1000, 2000, 100000000),
                     "module_label": outputModule, "dataset": datasetInfo}

        outputFile = loadTestReport.addOutputFile(outputModule, fileAttrs)
        Report.addRunInfoToFile(outputFile, runInfo)

    loadTestReport.persist("LoadTest%02d.pkl" % i)



#!/usr/bin/env python
"""Base class for integration test for request lifecycle"""
from __future__ import print_function

from WMCore.Services.ReqMgr.ReqMgr import ReqMgr
from WMCore.Services.WorkQueue.WorkQueue import WorkQueue
from WMCore.Cache.WMConfigCache import ConfigCache
from WMCore.WMBase import getTestBase

import nose
from nose.plugins.attrib import attr
import time
import os
import importlib
from functools import wraps

# decorator around tests - record errors
def recordException(fn):
    @wraps(fn)
    def wrapped(*args, **kwargs):
        try:
            return fn(*args, **kwargs)
        except:
            args[0].__class__._failure_detected = True
            raise
    return wrapped


class RequestLifeCycleBase_t(object):

    request = None
    request_name = None
    workqueue = None
    endpoint = os.environ.get('REQMGRBASEURL', 'https://localhost:8443')
    reqmgr = ReqMgr(endpoint + '/reqmgr2')
    team = 'TestTeam'
    _failure_detected = False

    @recordException
    def setUp(self):
        if self.__class__._failure_detected:
            raise nose.SkipTest
        # simple ping check - check reqmgr up
        tries = 0
        while True:
            try:
                if not self.__class__.request:
                    self.__class__.reqmgr.getTeam()
                break
            except:
                tries += 1
                if tries >= 3:
                    raise nose.SkipTest("Unable to contact reqmgr")
                time.sleep(15)

    def _configCacheId(self, label):
        """Return config cache id for given config label"""
        key, cert = self.__class__.reqmgr['requests'].getKeyCert()
        configCache = ConfigCache(self.__class__.endpoint + '/couchdb', 'reqmgr_config_cache', ckey = key, cert = cert)
        try:
            configCacheId = configCache.getIDFromLabel(label)
        except:
            configCacheId = None
        if configCacheId:
            return configCacheId
        # The following will fail if FWCore.ParameterSet not in PYTHONPATH
        from PSetTweaks.WMTweak import makeTweak
        configCache.createUserGroup('test', 'test')
        configDir = os.path.join(getTestBase(), '..', '..', 'test', 'data', 'configs')
        configCache.addConfig(os.path.join(configDir, label + '.py'))
        configCache.setLabel(label)
        configCache.setDescription(label)
        modSpecs = importlib.machinery.PathFinder().find_spec(label, [configDir])
        loadedConfig = modSpecs.loader.load_module()
        configCache.setPSetTweaks(makeTweak(loadedConfig.process).jsondictionary())
        configCache.save()
        return configCache.getIDFromLabel(label)

    def _convertLabelsToId(self, config):
        fields = ['ProcConfigCacheID', 'Skim1ConfigCacheID',
                  'StepOneConfigCacheID', 'ConfigCacheID']
        for field in fields:
            if config.get(field):
                config[field] = self._configCacheId(config[field])
        for field in ['Task1', 'Task2', 'Task3', 'Task4']:
            if config.get(field):
                config[field] = self._convertLabelsToId(config[field])
        return config

    @attr("lifecycle")
    @recordException
    def test05InjectConfigs(self):
        """Inject configs to cache"""
        self.__class__.requestParams = self._convertLabelsToId(self.__class__.requestParams)

    @attr("lifecycle")
    @recordException
    def test10InjectRequest(self):
        """Can inject a request"""
        self.__class__.requestParams.setdefault('RequestString', self.__class__.__name__)
        tries = 0
        while True:
            try:
                self.__class__.request = self.__class__.reqmgr.makeRequest(**self.__class__.requestParams)['WMCore.ReqMgr.DataStructs.Request']
                self.__class__.request_name = self.__class__.request['RequestName']
                break
            except:
                tries += 1
                if tries > 3:
                    raise
        self.assertTrue(self.__class__.request)
        self.assertTrue(self.__class__.request_name)
        print("Injected request %s" % self.__class__.request_name)
        self.__class__.request = self.__class__.reqmgr.getRequest(self.__class__.request_name)
        self.assertEqual(self.__class__.request['RequestStatus'], 'new')

    @attr("lifecycle")
    @recordException
    def test20ApproveRequest(self):
        """Approve request"""
        self.__class__.reqmgr.reportRequestStatus(self.__class__.request_name, 'assignment-approved')
        self.__class__.request = self.__class__.reqmgr.getRequest(self.__class__.request_name)
        self.assertEqual(self.__class__.request['RequestStatus'], 'assignment-approved')#

    @attr("lifecycle")
    @recordException
    def test30AssignRequest(self):
        """Assign request"""
        self.__class__.reqmgr.assign(self.__class__.request_name, self.__class__.team, "Testing", "v1",
                           MergedLFNBase='/store/temp', UnmergedLFNBase='/store/temp')
        self.__class__.request = self.reqmgr.getRequest(self.__class__.request_name)
        self.assertEqual(self.__class__.request['RequestStatus'], 'assigned')

    @attr("lifecycle")
    @recordException
    def test40WorkQueueAcquires(self):
        """WorkQueue picks up request"""
        if not self.__class__.request_name:
            raise nose.SkipTest
        start = time.time()
        while True:
            workqueue = self.reqmgr.getWorkQueue(request = self.__class__.request_name)
            if workqueue:
                self.__class__.workqueue = WorkQueue(workqueue[0])
                self.__class__.request = self.__class__.reqmgr.getRequest(self.__class__.request_name)
                self.assertTrue(self.__class__.request['RequestStatus'] in ('acquired', 'running'))
                request = [x for x in self.__class__.workqueue.getElementsCountAndJobsByWorkflow() if \
                    x == self.__class__.request_name]
                if [x for x in request if x['status'] in ('Available', 'Negotiating', 'Acquired', 'Running')]:
                    break
            if start + (60 * 20) < time.time():
                raise RuntimeError('timeout waiting for workqueue to acquire')
            time.sleep(15)

    @attr("lifecycle")
    @recordException
    def test50AgentAcquires(self):
        """Elements acquired by agent"""
        # skip if request already running
        self.__class__.request = self.__class__.reqmgr.getRequest(self.__class__.request_name)
        if self.__class__.request['RequestStatus'] == 'running':
            raise nose.SkipTest
        start = time.time()
        while True:
            request = [x for x in self.__class__.workqueue.getElementsCountAndJobsByWorkflow() if \
                        x == self.__class__.request_name]
            if [x for x in request if x['status'] in ('Acquired', 'Running')]:
                break
            if start + (60 * 20) < time.time():
                raise RuntimeError('timeout waiting for agent to acquire')
            time.sleep(15)
        self.assertTrue([x for x in request if x['status'] in ('Acquired', 'Running')])

    @attr("lifecycle")
    @recordException
    def test60RequestRunning(self):
        """Request running"""
        start = time.time()
        while True:
            request = [x for x in self.__class__.workqueue.getElementsCountAndJobsByWorkflow() if \
                    x == self.__class__.request_name]
            childQueue = [x for x in self.__class__.workqueue.getChildQueuesByRequest() if \
                    x['request_name'] == self.__class__.request_name]
            if request and 'Running' in [x['status'] for x in request]:
                self.assertTrue(childQueue, "Running but can't get child queue")
                break
            if start + (60 * 20) < time.time():
                raise RuntimeError('timeout waiting for request to run')
            time.sleep(15)

    @attr("lifecycle")
    @recordException
    def test70WorkQueueFinished(self):
        """Request completed in workqueue"""
        start = time.time()
        while True:
            request = [x for x in self.__class__.workqueue.getElementsCountAndJobsByWorkflow() if \
                    x == self.__class__.request_name]
            # request deleted from wq shortly after finishing, so may not appear here
            if not request or request == [x for x in request if x['status'] in ('Done', 'Failed', 'Canceled')]:
                break
            if start + (60 * 20) < time.time():
                raise RuntimeError('timeout waiting for request to finish')
            time.sleep(15)

    @attr("lifecycle")
    @recordException
    def test80RequestFinished(self):
        """Request completed"""
        start = time.time()
        while True:
            self.__class__.request = self.__class__.reqmgr.getRequest(self.__class__.request_name)
            if self.__class__.request['RequestStatus'] in ('completed', 'failed',
                                                           'aborted'):
                break
            if start + (60 * 20) < time.time():
                raise RuntimeError('timeout waiting for request to finish')
            time.sleep(15)

    @attr("lifecycle")
    @recordException
    def test90RequestCloseOut(self):
        """Closeout request"""
        self.reqmgr.reportRequestStatus(self.__class__.request_name, "closed-out")
        self.__class__.request = self.__class__.reqmgr.getRequest(self.__class__.request_name)
        self.assertEqual('closed-out', self.__class__.request['RequestStatus'])



#!/usr/bin/env python
"""
_List_

MySQL implementation of Subscription.List
"""

__all__ = []



from WMCore.Database.DBFormatter import DBFormatter

class List(DBFormatter):
    sql = "SELECT id FROM wmbs_subscription"

    def format(self, result):
        results = DBFormatter.format(self, result)

        subIDs = []
        for row in results:
            subIDs.append(row[0])

        return subIDs

    def execute(self, conn = None, transaction = False):
        result = self.dbi.processData(self.sql, conn = conn,
                                      transaction = transaction)
        return self.format(result)



#!/usr/bin/env python
"""
Unittests for the RequestInfoCollection WMStats module
"""
from __future__ import division, print_function

import unittest

from future.utils import viewvalues

from Utils.PythonVersion import PY3

from WMCore.Services.WMStats.DataStruct.RequestInfoCollection import (JobSummary, ProgressSummary,
                                                                      TaskInfo, RequestInfo)


class DummyTask(object):
    """
    Dummy object carrying some needed attributes for TaskInfo
    """

    def __init__(self, reqName, taskName, taskType):
        self.requestName = reqName
        self.taskName = taskName
        self.taskType = taskType
        self.jobSummary = {}

    def setJobSummary(self, jobSum):
        self.jobSummary = jobSum


class MyTestCase(unittest.TestCase):

    def setUp(self):
        if PY3:
            self.assertItemsEqual = self.assertCountEqual

    def testJobSummary(self):
        """some very basic unit tests for the JobSummary class"""
        jobSumKeys = ["success", "canceled", "transition", "queued",
                      "submitted", "failure", "cooloff", "paused"]
        jsonStatusKeys = ["success", "failure", "cooloff", "running",
                          "queued", "pending", "paused", "created"]

        # default object
        jobSumObj = JobSummary()
        self.assertItemsEqual(list(jobSumObj.jobStatus), jobSumKeys)
        self.assertItemsEqual(list(jobSumObj.getJSONStatus()), jsonStatusKeys)
        self.assertEqual(jobSumObj.getTotalJobs(), 0)
        self.assertEqual(sum(viewvalues(jobSumObj.getJSONStatus())), 0)

        # update with no content, thus, do nothing!
        jobSumObj.addJobStatusInfo({})
        self.assertItemsEqual(list(jobSumObj.jobStatus), jobSumKeys)
        self.assertItemsEqual(list(jobSumObj.getJSONStatus()), jsonStatusKeys)
        self.assertEqual(jobSumObj.getTotalJobs(), 0)
        self.assertEqual(sum(viewvalues(jobSumObj.getJSONStatus())), 0)

        # passing an invalid key/status
        jobSumObj.addJobStatusInfo({"bad_status_key": 10})
        self.assertItemsEqual(list(jobSumObj.jobStatus), jobSumKeys)
        self.assertEqual(jobSumObj.getTotalJobs(), 0)

        # updating a simple integer status
        jobSumObj.addJobStatusInfo({"success": 10})
        self.assertItemsEqual(list(jobSumObj.jobStatus), jobSumKeys)
        self.assertEqual(jobSumObj.getTotalJobs(), 10)
        self.assertEqual(jobSumObj.getSuccess(), 10)
        # getJSONStatus considers success jobs as success and as created,
        # thus double counting
        self.assertEqual(sum(viewvalues(jobSumObj.getJSONStatus())), 20)

        # updating a dictionary status
        jobSumObj.addJobStatusInfo({"submitted": {"pending": 2, "running": 4}})
        self.assertItemsEqual(list(jobSumObj.jobStatus), jobSumKeys)
        self.assertItemsEqual(list(jobSumObj.getJSONStatus()), jsonStatusKeys)
        self.assertEqual(jobSumObj.getTotalJobs(), 16)
        self.assertEqual(jobSumObj.getSuccess(), 10)
        # Submitted only considers first/retry states
        self.assertEqual(jobSumObj.getSubmitted(), 0)
        self.assertEqual(jobSumObj.getPending(), 2)
        self.assertEqual(jobSumObj.getRunning(), 4)
        self.assertEqual(sum(viewvalues(jobSumObj.getJSONStatus())), 32)

    def testProgressSummary(self):
        """some very basic unit tests for the ProgressSummary class"""
        progSumKeys = ["totalLumis", "events", "size"]

        # default object
        progSumObj = ProgressSummary()
        self.assertItemsEqual(list(progSumObj.getReport()), progSumKeys)
        self.assertEqual(sum(viewvalues(progSumObj.getReport())), 0)

        # update with no content, thus, do nothing!
        progSumObj.addProgressReport({})
        self.assertItemsEqual(list(progSumObj.getReport()), progSumKeys)
        self.assertEqual(sum(viewvalues(progSumObj.getReport())), 0)

        # passing an invalid key/status
        progSumObj.addProgressReport({"bad_status_key": 10})
        self.assertItemsEqual(list(progSumObj.getReport()), progSumKeys)
        self.assertEqual(sum(viewvalues(progSumObj.getReport())), 0)

        # now passing some valid information
        progSumObj.addProgressReport({"totalLumis": 10, "events": 1000})
        self.assertItemsEqual(list(progSumObj.getReport()), progSumKeys)
        self.assertEqual(sum(viewvalues(progSumObj.getReport())), 1010)

    def testTaskInfo(self):
        """some very basic unit tests for the TaskInfo class"""
        reqName = "test_request_name"
        taskName = "test_task_name"

        # default object
        taskSumObj = TaskInfo(reqName, taskName, {})
        self.assertEqual(taskSumObj.getRequestName(), reqName)
        self.assertEqual(taskSumObj.getTaskName(), taskName)
        self.assertEqual(taskSumObj.getTaskType(), "N/A")
        self.assertIsInstance(taskSumObj.getJobSummary(), JobSummary)
        self.assertFalse(taskSumObj.isTaskCompleted())

        # try to add an invalid task info object
        with self.assertRaises(Exception):
            taskSumObj.addTaskInfo("blah")

        # and again, with valid attributes but invalid values
        with self.assertRaises(Exception):
            dummyTask = DummyTask("req", "task", "type")
            taskSumObj.addTaskInfo(dummyTask)

        # now with all valid, but invalid JobSummary object
        dummyTask = DummyTask(reqName, taskName, "dummy_task_type")
        with self.assertRaises(AttributeError):
            taskSumObj.addTaskInfo(dummyTask)

        # now update it with some valid content
        dummyTask.setJobSummary(JobSummary())
        taskSumObj.addTaskInfo(dummyTask)
        self.assertFalse(taskSumObj.isTaskCompleted())

        # now complete this task
        # now update it with some valid content
        dummyTask.setJobSummary(JobSummary({"success": 10}))
        taskSumObj.addTaskInfo(dummyTask)
        self.assertTrue(taskSumObj.isTaskCompleted())

    def testRequestInfo(self):
        """some very basic unit tests for the RequestInfo class"""
        reqName = "test_request_name"
        taskName = "test_task_name"
        agentName = "test_agent_name"

        # default object
        defaultDict = {"RequestName": reqName,
                       "AgentJobInfo": {}}
        reqInfoObj = RequestInfo(defaultDict)
        self.assertItemsEqual(reqInfoObj.getTasks(), {})
        self.assertEqual(reqInfoObj.getTotalTopLevelJobs(), "N/A")
        self.assertEqual(reqInfoObj.getTotalInputLumis(), "N/A")
        self.assertEqual(reqInfoObj.getTotalInputEvents(), "N/A")
        self.assertEqual(reqInfoObj.getTotalTopLevelJobsInWMBS(), 0)
        self.assertItemsEqual(reqInfoObj.getJobSummaryByAgent(), {})
        self.assertItemsEqual(reqInfoObj.getTasksByAgent(), {})
        self.assertFalse(reqInfoObj.isWorkflowFinished())

        # with some useful content now
        defaultDict = {"RequestName": reqName,
                       "total_jobs": 10,
                       "input_lumis": 100,
                       "input_events": 1000,
                       "AgentJobInfo": {agentName: {"status": {"success": 10, "inWMBS": 10},
                                                    "tasks": {taskName: {"status": {"success": 10}}}}}}
        reqInfoObj = RequestInfo(defaultDict)
        self.assertItemsEqual(list(reqInfoObj.getTasks()), [taskName])
        self.assertEqual(reqInfoObj.getTotalTopLevelJobs(), 10)
        self.assertEqual(reqInfoObj.getTotalInputLumis(), 100)
        self.assertEqual(reqInfoObj.getTotalInputEvents(), 1000)
        self.assertEqual(reqInfoObj.getTotalTopLevelJobsInWMBS(), 10)
        self.assertIsInstance(reqInfoObj.getJobSummaryByAgent(), dict)
        self.assertTrue(agentName in reqInfoObj.getJobSummaryByAgent())
        self.assertIsInstance(reqInfoObj.getJobSummaryByAgent(agentName), JobSummary)
        self.assertIsInstance(reqInfoObj.getTasksByAgent(), dict)
        self.assertTrue(agentName in reqInfoObj.getTasksByAgent())
        self.assertIsInstance(reqInfoObj.getTasksByAgent(agentName), dict)
        self.assertTrue(taskName in reqInfoObj.getTasksByAgent(agentName))
        self.assertTrue(reqInfoObj.isWorkflowFinished())


if __name__ == '__main__':
    unittest.main()



"""
_DBSBufferDataset_t

Test module for the DBSBufferDataset object

Created on May 2, 2013

@author: dballest
"""
import os
import unittest
import threading

from WMComponent.DBS3Buffer.DBSBufferDataset import DBSBufferDataset

from WMCore.WMBase import getTestBase
from WMCore.WMSpec.WMWorkload import WMWorkloadHelper

from WMQuality.TestInit import TestInit

class DBSBufferDatasetTest(unittest.TestCase):
    """
    _DBSBufferDatasetTest_

    Collection of tests for the DBSBuffer object
    """
    def setUp(self):
        self.testInit = TestInit(__file__)
        self.testInit.setLogging()
        self.testInit.setDatabaseConnection()
        self.testInit.setSchema(customModules = ["WMComponent.DBS3Buffer"],
                                useDefault = False)

    def tearDown(self):
        self.testInit.clearDatabase()

    def testBasic(self):
        """
        _testBasic_

        Test the basic functions of the DBSBufferDataset,
        create, load, exists and also the ability
        to add subscriptions.
        """
        originalDataset = DBSBufferDataset(path = '/bogus/bogus/go')
        originalDataset.create()
        myThread = threading.currentThread()
        result = myThread.dbi.processData("SELECT id FROM dbsbuffer_dataset")[0].fetchall()
        self.assertEqual(originalDataset.exists(), result[0][0])
        duplicateDataset = DBSBufferDataset(path = '/bogus/bogus/go')
        duplicateDataset.create()
        self.assertEqual(originalDataset.exists(), duplicateDataset.exists())
        result = myThread.dbi.processData("SELECT COUNT(id) FROM dbsbuffer_dataset")[0].fetchall()
        self.assertEqual(result[0][0], 1)
        loadedDataset = DBSBufferDataset(path = '/bogus/bogus/go')
        loadedDataset.load()
        self.assertEqual(loadedDataset.exists(), originalDataset.exists())
        secondDataset = DBSBufferDataset(path = '/BogusPrimary/Run2012Z-PromptReco-v1/RECO')
        secondDataset.create()
        workload = WMWorkloadHelper()
        workload.load(os.path.join(getTestBase(), 'WMComponent_t/DBS3Buffer_t/specs/TestWorkload.pkl'))
        secondDataset.addSubscription(workload.getSubscriptionInformation()['/BogusPrimary/Run2012Z-PromptReco-v1/RECO'])
        secondDataset.addSubscription(workload.getSubscriptionInformation()['/BogusPrimary/Run2012Z-PromptReco-v1/RECO'])
        self.assertEqual(len(secondDataset['subscriptions']), 3)
        result = myThread.dbi.processData("SELECT COUNT(id) FROM dbsbuffer_dataset_subscription")[0].fetchall()
        self.assertEqual(result[0][0], 3)
        return

if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
"""
_GetGroupsByJobState_

MySQL implementation of JobGroup.GetGroupsByJobState
"""




from WMCore.Database.DBFormatter import DBFormatter

class GetGroupsByJobState(DBFormatter):
    sql = """SELECT DISTINCT wmbs_jobgroup.id FROM wmbs_jobgroup
               INNER JOIN wmbs_job ON
                 wmbs_jobgroup.id = wmbs_job.jobgroup
               INNER JOIN wmbs_job_state ON
                 wmbs_job.state = wmbs_job_state.id
              WHERE wmbs_job_state.name = :job_state
              AND wmbs_jobgroup.subscription >= :minsub"""

    def format(self, results):
        """
        _format_

        Format the jobgroup ids into a single list.
        """
        results = DBFormatter.format(self, results)

        jobGroupList = []
        for result in results:
            jobGroupList.append(result[0])

        return jobGroupList

    def execute(self, jobState = None, minSub = 0, conn = None, transaction = False):
        result = self.dbi.processData(self.sql, {"job_state": jobState, "minsub": minSub},
                                      conn = conn, transaction = transaction)
        return self.format(result)



#!/usr/bin/env python
"""
This script is meant to be used when an agent has - apparently - fully
drained, without any jobs in condor.

It checks many things, most - if not all - already checked by the DrainStatusPoller
thread as well. It checks status of:
* workflows
* subscriptions
* jobs (in condor, wmbs and bossair
* DBS injection
* PhEDEx injection
* etc

NOTE: you need to source the agent environment:
source apps/wmagent/etc/profile.d/init.sh
 """


from builtins import range, str, bytes

import argparse
import logging
import os
import socket
import sys
import threading
import time
from pprint import pprint, pformat
from time import gmtime, strftime

import htcondor as condor

from Utils.IteratorTools import flattenList
from WMCore.Configuration import loadConfigurationFile
from WMCore.Database.DBFormatter import DBFormatter
from WMCore.Services.RequestDB.RequestDBReader import RequestDBReader
from WMCore.WMInit import connectToDB
from WMCore.WorkQueue.WorkQueueBackend import WorkQueueBackend

jobCountByState = """
    select wmbs_job_state.name, count(*) AS count
      from wmbs_job
      join wmbs_job_state on (wmbs_job.state = wmbs_job_state.id)
      group by wmbs_job.state, wmbs_job_state.name
    """

incompleteWfs = "SELECT DISTINCT name FROM dbsbuffer_workflow WHERE completed = 0"

knownWorkflows = "SELECT DISTINCT name from wmbs_workflow"

workflowsNotInjected = "select distinct name from wmbs_workflow where injected = 0"

unfinishedSubscriptions = """
   select distinct wmbs_workflow.name AS wfName
     FROM wmbs_subscription
     INNER JOIN wmbs_fileset ON wmbs_subscription.fileset = wmbs_fileset.id
     INNER JOIN wmbs_workflow ON wmbs_workflow.id = wmbs_subscription.workflow
     where wmbs_subscription.finished = 0 ORDER BY wmbs_workflow.name
   """

filesAvailWMBS = "select subscription,count(*) from wmbs_sub_files_available group by subscription"

filesAcqWMBS = "select subscription,count(*) from wmbs_sub_files_acquired group by subscription"

blocksOpenDBS = "SELECT * FROM dbsbuffer_block WHERE status!='Closed'"

filesNotInDBS = "SELECT lfn from dbsbuffer_file where status = 'NOTUPLOADED'"

filesNotInPhedex = """
    SELECT lfn FROM dbsbuffer_file
      WHERE in_phedex=0
      AND block_id IS NOT NULL
      AND lfn NOT LIKE '%%unmerged%%'
      AND lfn NOT LIKE 'MCFakeFile%%'
      AND lfn NOT LIKE '%%BACKFILL%%'
      AND lfn NOT LIKE '/store/backfill/%%'
      AND lfn NOT LIKE '/store/user%%'
    """
filesNotInPhedexNull = """
    SELECT lfn FROM dbsbuffer_file
      WHERE in_phedex=0
      AND block_id IS NULL
      AND lfn NOT LIKE '%%unmerged%%'
      AND lfn NOT LIKE 'MCFakeFile%%'
      AND lfn NOT LIKE '%%BACKFILL%%'
      AND lfn NOT LIKE '/store/backfill/%%'
      AND lfn NOT LIKE '/store/user%%'
    """

workflowsExecuting = """
    SELECT wmbs_workflow.name AS name, count(wmbs_job.id) AS count
    FROM wmbs_job
    INNER JOIN wmbs_jobgroup ON wmbs_job.jobgroup = wmbs_jobgroup.id
    INNER JOIN wmbs_subscription ON wmbs_jobgroup.subscription = wmbs_subscription.id
    INNER JOIN wmbs_sub_types ON wmbs_subscription.subtype = wmbs_sub_types.id
    INNER JOIN wmbs_job_state ON wmbs_job.state = wmbs_job_state.id
    INNER JOIN wmbs_workflow ON wmbs_subscription.workflow = wmbs_workflow.id
    WHERE wmbs_job_state.name = 'executing' GROUP BY wmbs_workflow.name
    """


def printWfStatus(wfs, workflowsDict):
    """
    Given a list of request names, print their RequestStatus from the short dictionary.
    """
    for wf in wfs:
        print("%-125s\t%s" % (wf, workflowsDict[wf]['RequestStatus']))

    return


def lfn2dset(lfns):
    """
    Convert a LFN into a dataset name
    """
    if isinstance(lfns, (str, bytes)):
        lfns = [lfns]

    listDsets = set()
    for lfn in lfns:
        toks = lfn.split('/')[3:]
        dset = "/%s/%s-%s/%s" % (toks[1], toks[0], toks[3], toks[2])
        listDsets.add(dset)

    return listDsets


def filterKeys(schema):
    """
    Saves only input and output dataset info
    """
    newSchema = {}
    newSchema['RequestStatus'] = schema.get('RequestStatus', "")
    newSchema['InputDataset'] = schema.get('InputDataset', "")
    newSchema['OutputDatasets'] = schema.get('OutputDatasets', "")

    if schema['RequestType'] in ['StepChain', 'TaskChain']:
        joker = schema['RequestType'].split('Chain')[0]
        numInnerDicts = schema[schema['RequestType']]
        # and now we look for this key in each Task/Step
        for i in range(1, numInnerDicts + 1):
            innerName = "%s%s" % (joker, i)
            if 'InputDataset' in schema[innerName]:
                newSchema['InputDataset'] = schema[innerName]['InputDataset']
    return newSchema


def getDsetAndWf(lfns, wfsDict):
    """
    Given a list of lfns (files not injected in DBS/TMDB), get their dataset
    name and compare against the outputdatasets of workflows known by this agent.

    If there is no match, then the workflow has been archived already and there
    is nothing to worry about.

    If the workflow is known, then we may need to recover these files.
    """
    if not lfns:
        return
    uniqLfns = set([lfn.rsplit('/', 2)[0] for lfn in lfns])
    uniqDsets = lfn2dset(uniqLfns)
    print("==> Which maps to %d unique datasets:\n%s" % (len(uniqDsets), pformat(uniqDsets)))

    match = []
    for dset in uniqDsets:
        for wf, values in wfsDict.items():
            if dset in values['OutputDatasets']:
                match.append((wf, values['RequestStatus'], dset))
    if match:
        print("... that were produced by the following workflow, status and the dataset itself:\n")
        pprint(match)
    else:
        print("... that were NOT produced by any agent-known workflow. OR, the wfs are gone already.\n")

    return


def fetchWorkflowsSpec(config, listOfWfs):
    """
    Fetch the workload of a list of workflows. Filter out only a few
    usefull keys
    """
    if isinstance(listOfWfs, (str, bytes)):
        listOfWfs = [listOfWfs]

    wfDBReader = RequestDBReader(config.AnalyticsDataCollector.centralRequestDBURL,
                                 couchapp=config.AnalyticsDataCollector.RequestCouchApp)
    tempWfs = wfDBReader.getRequestByNames(listOfWfs, True)

    wfShortDict = {}
    for wf in listOfWfs:
        wfShortDict[wf] = filterKeys(tempWfs[wf])

    return wfShortDict


def getCondorJobs():
    """
    Retrieve jobs from condor.
    :return: amount of jobs ordered by JobStatus and workflow
    """
    jobDict = {}
    schedd = condor.Schedd()
    jobs = schedd.xquery('WMAgent_AgentName == "WMAgent"', ['WMAgent_RequestName', 'JobStatus'])
    for job in jobs:
        jobStatus = job['JobStatus']
        jobDict.setdefault(jobStatus, {})
        jobDict[jobStatus].setdefault(job['WMAgent_RequestName'], 0)
        jobDict[jobStatus][job['WMAgent_RequestName']] += 1
    return jobDict


def checkLocalWQStatus(config, status):
    """
    Given a WorkQueueElement status, query local workqueue and workqueue_inbox
    database for all elements in a given status and that were acquired by this agent.
    """
    backend = WorkQueueBackend(config.WorkQueueManager.couchurl)

    for db in ("workqueue", "workqueue_inbox"):
        if db == "workqueue":
            elements = backend.getElements(status=status)
        else:
            elements = backend.getInboxElements(status=status)

        for elem in elements:
            updatedIn = time.strftime("%a, %d %b %Y %H:%M:%S %Z", time.localtime(float(elem.updatetime)))
            print("id: %s\tRequestName: %s\tStatus: %s\t\tUpdatedIn: %s" % (
                elem.id, elem['RequestName'], elem['Status'], updatedIn))
        print("Elements matching the criteria (%s, %s) are: %d" % (status, db, len(elements)))
    return


def checkGlobalWQStatus(config, status):
    """
    Given a WorkQueueElement status, query central workqueue database for
    all elements in a given status and that were acquired by this agent.
    """
    agentUrl = "http://" + socket.gethostname() + ":5984"

    backend = WorkQueueBackend(config.WorkloadSummary.couchurl)
    elements = backend.getElements(status=status, ChildQueueUrl=agentUrl)

    for elem in elements:
        updatedIn = time.strftime("%a, %d %b %Y %H:%M:%S %Z", time.localtime(float(elem.updatetime)))
        print("id: %s\tRequestName: %s\tStatus: %s\t\tUpdatedIn: %s" % (
            elem.id, elem['RequestName'], elem['Status'], updatedIn))
    print("Elements matching the criteria (%s, %s) are: %d" % (status, agentUrl, len(elements)))
    return


def createElementsSummary(elements, queueUrl):
    """
    Print the local couchdb situation based on the WQE status
    """
    summary = {'numberOfElements': len(elements), 'queueURL': queueUrl}
    for elem in elements:
        summary.setdefault(elem['Status'], 0)
        summary[elem['Status']] += 1
    if set(summary.keys()) == set(['numberOfElements', 'queueURL', 'Done']):
        print("Done in: %s" % summary['queueURL'])
    else:
        print(summary)


def getWMBSInfo(config):
    connectToDB()
    myThread = threading.currentThread()
    formatter = DBFormatter(logging, myThread.dbi)

    workflows = formatter.formatDict(myThread.dbi.processData(knownWorkflows))
    workflows = [wf['name'] for wf in workflows]
    print("\n*** WORKFLOWS: found %d distinct workflows in this agent." % len(workflows))
    workflowsDict = fetchWorkflowsSpec(config, workflows)
    printWfStatus(workflows, workflowsDict)

    for st in ('Available', 'Negotiating', 'Acquired', 'Running'):
        print("\n*** WORKQUEUE: elements still marked as %s in LQ workqueue / workqueue_inbox." % st)
        checkLocalWQStatus(config, st)

    for st in ("Acquired", "Running"):
        print("\n*** WORKQUEUE: elements still marked as %s in GQ workqueue." % st)
        checkGlobalWQStatus(config, st)

    workflows = formatter.formatDict(myThread.dbi.processData(incompleteWfs))
    workflows = [wf['name'] for wf in workflows]
    print("\n*** WORKFLOWS: there are %d distinct workflows not completed." % len(workflows))
    printWfStatus(workflows, workflowsDict)

    wfsNotInjected = flattenList(formatter.format(myThread.dbi.processData(workflowsNotInjected)))
    print("\n*** WORKFLOWS: found %d workflows not fully injected." % len(wfsNotInjected))
    printWfStatus(wfsNotInjected, workflowsDict)

    jobsByState = formatter.formatDict(myThread.dbi.processData(jobCountByState))
    print("\n*** WMBS: amount of wmbs jobs in each status:\n%s" % jobsByState)
    # IF we have executing jobs in wmbs and nothing in condor, then investigate the wfs
    if 'executing' in [item['name'] for item in jobsByState]:
        wfsJobCount = formatter.formatDict(myThread.dbi.processData(workflowsExecuting))
        print("\n*** WMBS: %d workflows with executing jobs in wmbs:" % len(wfsJobCount))
        workflows = [wf['name'] for wf in wfsJobCount]
        printWfStatus(workflows, workflowsDict)

    unfinishedSubs = formatter.formatDict(myThread.dbi.processData(unfinishedSubscriptions))
    unfinishedSubs = [wf['wfname'] for wf in unfinishedSubs]
    print("\n*** SUBSCRIPTIONS: subscriptions not finished: %d" % len(unfinishedSubs))
    printWfStatus(unfinishedSubs, workflowsDict)

    filesAvailable = formatter.formatDict(myThread.dbi.processData(filesAvailWMBS))
    print(
        "\n*** SUBSCRIPTIONS: found %d files available in WMBS (waiting for job creation):\n%s" % (len(filesAvailable),
                                                                                                   filesAvailable))

    filesAcquired = formatter.formatDict(myThread.dbi.processData(filesAcqWMBS))
    print(
        "\n*** SUBSCRIPTIONS: found %d files acquired in WMBS (waiting for jobs to finish):\n%s" % (len(filesAcquired),
                                                                                                    filesAcquired))

    blocksopenDBS = formatter.formatDict(myThread.dbi.processData(blocksOpenDBS))
    print("\n*** DBS: found %d blocks open in DBS." % len(blocksopenDBS), end="")
    print(" Printing the first 20 blocks only:\n%s" % blocksopenDBS[:20])

    filesnotinDBS = flattenList(formatter.format(myThread.dbi.processData(filesNotInDBS)))
    print("\n*** DBS: found %d files not uploaded to DBS.\n" % len(filesnotinDBS))
    getDsetAndWf(filesnotinDBS, workflowsDict)

    filesnotinPhedex = flattenList(formatter.format(myThread.dbi.processData(filesNotInPhedex)))
    print("\n*** PHEDEX: found %d files not injected in PhEDEx, with valid block id (recoverable)." % len(
        filesnotinPhedex))
    getDsetAndWf(filesnotinPhedex, workflowsDict)

    filesnotinPhedexNull = flattenList(formatter.format(myThread.dbi.processData(filesNotInPhedexNull)))
    print("\n*** PHEDEX: found %d files not injected in PhEDEx, with valid block id (unrecoverable)." % len(
        filesnotinPhedexNull))
    getDsetAndWf(filesnotinPhedexNull, workflowsDict)


def parseArgs():
    """
    Well, parse the arguments passed in the command line :)
    """
    parser = argparse.ArgumentParser(description="Does a bunch of draining checks")
    parser.add_argument('-t', '--twiki', action='store_true', default=False,
                        help='Use it to get an output that can be directly pasted in a twiki')
    args = parser.parse_args()
    return args


def main():
    """
    Retrieve the following information from the agent:
      1. number of jobs in condor
      2. list of distinct workflows in wmbs_workflow (and their status in reqmgr2)
      3. amount of wmbs jobs in each status
      4. list of workflows not fully injected
      5. list of subscriptions not finished
      6. amount of files available in wmbs
      7. amount of files acquired in wmbs
      8. list of blocks not closed in phedex/dbs
      9. list of files not uploaded to dbs
      10. list of files not injected into phedex, with parent block
      11. list of files not injected into phedex, without parent block
    """
    args = parseArgs()

    twiki = ('<pre>', '</pre>') if args.twiki else ('', '')
    print(twiki[0])

    os.environ['WMAGENT_CONFIG'] = '/data/srv/wmagent/current/config/wmagent/config.py'
    config = loadConfigurationFile(os.environ["WMAGENT_CONFIG"])

    gmtTimeNow = strftime("%a, %d %b %Y %H:%M:%S +0000", gmtime())
    print("Drain check time: %s\n" % gmtTimeNow)

    print("*** Amount of jobs in condor per workflow, sorted by condor job status:")
    pprint(getCondorJobs())

    getWMBSInfo(config)

    print(twiki[1])
    print("I'm done!")
    sys.exit(0)


if __name__ == '__main__':
    sys.exit(main())



"""
Created on Nov. 7, 2023 by Duong Nguyen
"""
import os
import unittest

from WMCore.Storage.Registry import retrieveStageOutImpl
from WMCore.Storage.SiteLocalConfig import stageOutStr
from WMCore.Storage.StageInMgr import StageInMgr
from WMCore.Storage.StageOutError import StageOutFailure
from WMCore.Storage.StageOutMgr import searchRFC


class StageInMgrTest(StageInMgr):
    def stageIn(self, lfn, stageOut_rfc=None):
        """
        Given the lfn and a pair of stage out and corresponding Rucio file catalog, stageOut_rfc, or override configuration invoke the stage in
        If use override configuration self.overrideConf should contain:
        command - the stage out impl plugin name to be used
        option - the option values to be passed to that command (None is allowed)
        lfn-prefix - the LFN prefix to generate the PFN
        phedex-node - the Name of the PNN to which the file is being xferred
        :param lfn: logical file name
        :param stageOut_rfc: a pair of stage out attributes and corresponding Rucio file catalog
        """
        localPfn = os.path.join(os.getcwd(), os.path.basename(lfn))
        if self.override:
            pnn = self.overrideConf['phedex-node']
            command = self.overrideConf['command']
            options = self.overrideConf['option']
            if self.overrideConf['lfn-prefix'] is None:
                msg = "Unable to match lfn to pfn during stage in because override lfn-prefix is None: \n %s" % lfn
                raise StageOutFailure(msg, LFN=lfn)
            pfn = "%s%s" % (self.overrideConf['lfn-prefix'], lfn)
            protocol = command
        else:
            if not stageOut_rfc:
                msg = "Can not perform stage in for this lfn because of missing information (stageOut_rfc is None or empty): \n %s" % lfn
                raise StageOutFailure(msg, LFN=lfn)
            pnn = stageOut_rfc[0]['phedex-node']
            command = stageOut_rfc[0]['command']
            options = stageOut_rfc[0]['option']
            pfn = searchRFC(stageOut_rfc[1], lfn)
            protocol = stageOut_rfc[1].preferredProtocol

        if pfn == None:
            msg = "Unable to match lfn to pfn: \n  %s" % lfn
            raise StageOutFailure(msg, LFN=lfn, StageOut=stageOutStr(stageOut_rfc[0]))
        try:
            impl = retrieveStageOutImpl(command, stagein=True)
        except Exception as ex:
            msg = "Unable to retrieve impl for local stage in:\n"
            msg += "Error retrieving StageOutImpl for command named: %s\n" % (command,)
            raise StageOutFailure(msg, Command=command,
                                  LFN=lfn, ExceptionDetail=str(ex))
        impl.numRetries = self.numberOfRetries
        impl.retryPause = self.retryPauseTime

        try:
            pass
        except Exception as ex:
            msg = "Failure for stage in:\n"
            msg += str(ex)
            raise StageOutFailure(msg, Command=command, Protocol=protocol,
                                  LFN=lfn, InputPFN=localPfn, TargetPFN=pfn)
        return localPfn


class StageInMgrUnitTest(unittest.TestCase):

    def setUp(self):
        # shut up SiteLocalConfig
        os.putenv('CMS_PATH', os.getcwd())
        os.putenv('SITECONFIG_PATH', os.getcwd())

    def testStageInMgr(self):
        os.environ['SITECONFIG_PATH'] = '/cvmfs/cms.cern.ch/SITECONF/T1_US_FNAL'
        stageInMgr = StageInMgrTest()
        # keep using 'phedex-node' for overrideParams (to be compatible with the whole DMWM structure?)
        stageInMgr_override = StageInMgrTest(
            **{"command": "gfal2", "phedex-node": "T1_US_FNAL_Disk", "lfn-prefix": "root://abc/xyz"})
        fileToStage = {'LFN': '/store/abc/xyz.root', 'PFN': ''}
        stageInMgr(**fileToStage)
        stageInMgr_override(**fileToStage)
        return


if __name__ == "__main__":
    # import sys;sys.argv = ['', 'Test.testName']
    unittest.main()



#!/usr/bin/env python
"""
_WMExecutionFailure_

Error class to generate when an executor implementation encounters a problem

"""



from WMCore.WMException import WMException


class WMExecutionFailure(WMException):


    def __init__(self, code, name, detail):
        WMException.__init__(self, detail, code)
        self.code = code
        self.name = name
        self.detail = detail



#!/usr/bin/env python
# encoding: utf-8
"""
CMSSW_t.py

Test the CMSSW Executor and its associated Diagnostics object
under different simulated circumstances.

Created by Dave Evans on 2010-03-15.
Copyright (c) 2010 Fermilab. All rights reserved.
"""

import inspect
import os
import shutil
import subprocess
import sys
import unittest
from resource import getrusage, RUSAGE_SELF, RUSAGE_CHILDREN

import WMCore_t.WMSpec_t.Steps_t as ModuleLocator

from WMCore.DataStructs.Job import Job
from WMCore.FwkJobReport.Report import Report
from WMCore.WMBase import getTestBase
from WMCore.WMSpec.Makers.TaskMaker import TaskMaker
from WMCore.WMSpec.Steps import StepFactory
from WMCore.WMSpec.Steps.Executors.CMSSW import CMSSW as CMSSWExecutor
from WMCore.WMSpec.Steps.Templates.CMSSW import CMSSW as CMSSWTemplate
from WMCore.WMSpec.Steps.WMExecutionFailure import WMExecutionFailure
from WMCore.WMSpec.WMWorkload import newWorkload
from WMQuality.TestInit import TestInit


class CMSSW_t(unittest.TestCase):
    def setUp(self):
        """
        _setUp_

        Build a testing environment similar to a WN
        """
        self.testInit = TestInit(__file__)
        self.testDir = self.testInit.generateWorkDir()

        # Build a workload/task/step with the basic required information
        self.workload = newWorkload("UnitTests")
        self.task = self.workload.newTask("CMSSWExecutor")
        stepHelper = self.task.makeStep("ExecutorTest")
        self.step = stepHelper.data
        template = CMSSWTemplate()
        template(self.step)
        self.helper = template.helper(self.step)
        self.step.application.setup.scramCommand = "scramulator.py"
        self.step.application.command.executable = "cmsRun.py"
        self.step.application.setup.scramProject = "CMSSW"
        self.step.application.setup.scramArch = "slc5_ia32_gcc434"
        self.step.application.setup.cmsswVersion = "CMSSW_X_Y_Z"
        self.step.application.setup.softwareEnvironment = "echo \"Software Setup...\";"
        self.step.output.jobReport = "FrameworkJobReport.xml"
        self.helper.addOutputModule("outputRECORECO", primaryDataset="Bogus",
                                    processedDataset="Test-Era-v1",
                                    dataTier="DATA")
        self.helper.addOutputModule("outputALCARECORECO", primaryDataset="Bogus",
                                    processedDataset="Test-Era-v1",
                                    dataTier="DATA")
        self.helper.setGlobalTag("Bogus")
        taskMaker = TaskMaker(self.workload, self.testDir)
        taskMaker.skipSubscription = True
        taskMaker.processWorkload()

        # Build the TaskSpace/StepSpace
        self.sandboxDir = os.path.join(self.testDir, "UnitTests")
        self.task.build(self.testDir)
        sys.path.append(self.testDir)
        sys.path.append(self.sandboxDir)

        # Copy the files that cmsRun would have generated in the step space
        open(os.path.join(self.step.builder.workingDir, "outputRECORECO.root"), "w").close()
        open(os.path.join(self.step.builder.workingDir, "outputALCARECORECO.root"), "w").close()
        shutil.copy(os.path.join(getTestBase(),
                                 "WMCore_t/FwkJobReport_t/CMSSWProcessingReport.xml"),
                    os.path.join(self.step.builder.workingDir, "FrameworkJobReport.xml"))

        # Create a job
        self.job = Job(name="/UnitTest/CMSSWExecutor/ExecutorTest-test-job")
        self.job["id"] = 1

        # Set the PATH
        binDir = inspect.getsourcefile(ModuleLocator)
        binDir = binDir.replace("__init__.py", "bin")

        if not binDir in os.environ['PATH']:
            os.environ['PATH'] = "%s:%s" % (os.environ['PATH'], binDir)

        self.oldCwd = os.getcwd()

    def tearDown(self):
        """
        _tearDown_

        Clean up
        """
        self.testInit.delWorkDir()
        sys.path.remove(self.testDir)
        sys.path.remove(self.sandboxDir)
        try:
            sys.modules.pop("WMTaskSpace")
            sys.modules.pop("WMTaskSpace.ExecutorTest")
            sys.modules.pop("WMTaskSpace.ExecutorTest.WMCore")
            sys.modules.pop("WMSandbox")
            sys.modules.pop("WMSandbox.CMSSWExecutor")
            sys.modules.pop("WMSandbox.CMSSWExecutor.ExecutorTest")
        except KeyError:
            return

        return

    def testA_Execute(self):
        """
        _Execute_

        Test the full path, including execute.
        """
        try:
            os.chdir(self.step.builder.workingDir)
            executor = CMSSWExecutor()
            executor.initialise(self.step, self.job)
            executor.pre()
            # Remove the preScripts, as it is not needed for this test
            executor.step.runtime.preScripts.remove("SetupCMSSWPset")
            executor.execute()
            executor.post()
            # Check that there were no errors
            self.assertEqual(0, executor.report.getExitCode())
            # Check that we processed the XML
            self.assertEqual(2, len(executor.report.getAllFiles()))
            # Check that the executable was really executed
            self.assertTrue(os.path.isfile(os.path.join(self.step.builder.workingDir, "BogusFile.txt")))
        except Exception as ex:
            self.fail("Failure encountered, %s" % str(ex))
        finally:
            os.chdir(self.oldCwd)
        return

    def testB_ExecuteNonZeroExit(self):
        """
        _ExecuteNonZeroExit_

        Test the execution of a script
        which exits with non-zero code.
        """
        self.step.application.command.executable = "brokenCmsRun.py"
        shutil.copy(os.path.join(getTestBase(),
                                 "WMCore_t/FwkJobReport_t/CMSSWFailReport.xml"),
                    os.path.join(self.step.builder.workingDir, "FrameworkJobReport.xml"))
        try:
            os.chdir(self.step.builder.workingDir)
            executor = StepFactory.getStepExecutor("CMSSW")
            executor.initialise(self.step, self.job)
            executor.pre()
            executor.step.runtime.preScripts.remove("SetupCMSSWPset")
            try:
                executor.execute()
                self.fail("An exception should have been raised")
            except WMExecutionFailure as ex:
                executor.diagnostic(ex.code, executor, ExceptionInstance=ex)
                self.assertEqual(8001, executor.report.getExitCode())
                report = Report()
                report.load("Report.pkl")
                self.assertEqual(8001, report.getExitCode())
        except Exception as ex:
            self.fail("Failure encountered, %s" % str(ex))
        finally:
            os.chdir(self.oldCwd)
        return

    def testC_ExecuteSegfault(self):
        """
        _ExecuteSegfault_

        Test the execution of a script
        which raises a ABRT signal which
        is the normal CMSSW response
        to a SEGFAULT.
        """
        self.step.application.command.executable = "test.sh"
        # CMSSW leaves an empty FWJR when a SEGFAULT is present
        open(os.path.join(self.step.builder.workingDir, "FrameworkJobReport.xml"), "w").close()
        try:
            os.chdir(self.step.builder.workingDir)
            executor = StepFactory.getStepExecutor("CMSSW")
            executor.initialise(self.step, self.job)
            executor.pre()
            executor.step.runtime.preScripts.remove("SetupCMSSWPset")
            try:
                executor.execute()
                self.fail("An exception should have been raised")
            except WMExecutionFailure as ex:
                executor.diagnostic(ex.code, executor, ExceptionInstance=ex)
                #self.assertEqual(50115, executor.report.getExitCode())
                self.assertEqual(134, executor.report.getExitCode())
                report = Report()
                report.load("Report.pkl")
                self.assertEqual(134, report.getExitCode())
        except Exception as ex:
            self.fail("Failure encountered, %s" % str(ex))
        finally:
            os.chdir(self.oldCwd)
        return

    def testD_ExecuteNoOutput(self):
        """
        _ExecuteNoOutput_

        Test what happens when no output is produced,
        the proper error should be included.
        """
        self.step.application.command.executable = "cmsRun.py"
        shutil.copy(os.path.join(getTestBase(),
                                 "WMCore_t/FwkJobReport_t/CMSSWSkippedAll.xml"),
                    os.path.join(self.step.builder.workingDir, "FrameworkJobReport.xml"))
        try:
            os.chdir(self.step.builder.workingDir)
            executor = StepFactory.getStepExecutor("CMSSW")
            executor.initialise(self.step, self.job)
            executor.pre()
            executor.step.runtime.preScripts.remove("SetupCMSSWPset")
            executor.execute()
            executor.post()
            self.assertEqual(60450, executor.report.getExitCode())
        except Exception as ex:
            self.fail("Failure encountered, %s" % str(ex))
        finally:
            os.chdir(self.oldCwd)
        return

    def testSubprocessTime(self):
        """
        Test logic of calculating subprocess user/sys times
        """
        # take snapshot of user and sys time before we call subprocess
        userTime0 = getrusage(RUSAGE_CHILDREN).ru_utime
        sysTime0 = getrusage(RUSAGE_CHILDREN).ru_stime

        # call subprocess command
        args = ["sleep", "1"]
        proc = subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        childPid, exitStatus, resource = os.wait4(proc.pid, os.P_PID)
        stdout, stderr = proc.communicate()
        returnCode = proc.returncode
        self.assertTrue(stdout == b'')
        self.assertTrue(stderr == b'')
        self.assertTrue(childPid >= 0)
        self.assertTrue(exitStatus == 0)
        self.assertTrue(returnCode == 0)
        self.assertTrue(resource.ru_utime > 0)
        self.assertTrue(resource.ru_stime > 0)

        # calculate user and sys time of subprocess by substructing relevant
        # parts take before we spawn the subprocess
        userTime1 = getrusage(RUSAGE_CHILDREN).ru_utime
        sysTime1 = getrusage(RUSAGE_CHILDREN).ru_stime

        self.assertTrue(userTime1 >= userTime0)
        self.assertTrue(sysTime1 >= sysTime0)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
"""
Unittests for the Rucio wrapper PyCurl-based module
"""
from __future__ import division, print_function
from builtins import str

import unittest

from Utils.PythonVersion import PY3

from WMCore.MicroService.Tools.PycurlRucio import (getRucioToken, parseNewLineJson,
                                                   getPileupContainerSizesRucio, listReplicationRules,
                                                   getBlocksAndSizeRucio, stringDateToEpoch)

CONT1 = "/NoBPTX/Run2018D-12Nov2019_UL2018-v1/MINIAOD"
CONT2 = "/NoBPTX/Run2016F-23Sep2016-v1/DQMIO"
CONT3 = "/RelValTTbar_14TeV/CMSSW_11_2_0_pre8-112X_mcRun3_2024_realistic_v10_forTrk-v1/GEN-SIM"
# Pileup container
PU_CONT = "/Neutrino_E-10_gun/RunIISummer20ULPrePremix-UL16_106X_mcRun2_asymptotic_v13-v1/PREMIX"
PU_CONT_BLK = "/Neutrino_E-10_gun/RunIISummer20ULPrePremix-UL16_106X_mcRun2_asymptotic_v13-v1/PREMIX#00287791-198c-4000-aafa-a796878d51bb"


class PycurlRucioTests(unittest.TestCase):
    """Test pycurl_manager module"""

    def setUp(self):
        "initialization"
        self.rucioUrl = "http://cms-rucio-int.cern.ch"
        self.rucioAuthUrl = "https://cms-rucio-auth-int.cern.ch"
        self.rucioAccount = "wma_test"
        self.rucioScope = "cms"
        #self.rucioToken, self.tokenValidity = getRucioToken(self.rucioAuthUrl, self.rucioAccount)
        self.badDID = "/wrong/did/name"
        if PY3:
            self.assertItemsEqual = self.assertCountEqual

    def testParseNewLineJson(self):
        """
        Test the parseNewLineJson function
        """
        validate = {"1": {"first_item": 1},
                    "2": {"second_item": "2nd"},
                    "3": {"third_item": None}}
        dataStream = '{"first_item": 1}\n{"second_item": "2nd"}\n{"third_item": null}\n'
        for num, item in enumerate(parseNewLineJson(dataStream)):
            self.assertItemsEqual(item, validate[str(num + 1)])

    def testGetPileupContainerSizesRucio(self):
        """
        Test the getPileupContainerSizesRucio function, which fetches the container
        total bytes.
        """
        self.rucioToken, self.tokenValidity = getRucioToken(self.rucioAuthUrl, self.rucioAccount)
        # Test 1: no DIDs provided as input
        resp = getPileupContainerSizesRucio([], self.rucioUrl,
                                            self.rucioToken, scope=self.rucioScope)
        self.assertEqual(resp, {})

        # Test 2: multiple valid/invalid DIDs provided as input
        containers = [CONT1, PU_CONT, self.badDID]
        resp = getPileupContainerSizesRucio(containers, self.rucioUrl,
                                            self.rucioToken, scope=self.rucioScope)
        self.assertTrue(len(resp) == 3)
        self.assertTrue(resp[PU_CONT] > 0)
        self.assertIsNone(resp[self.badDID])

    def testListReplicationRules(self):
        """
        Test the listReplicationRules function, which fetches the container
        rules a return a list of RSEs key'ed by the container name.
        """
        self.rucioToken, self.tokenValidity = getRucioToken(self.rucioAuthUrl, self.rucioAccount)
        resp = listReplicationRules([], self.rucioAccount, grouping="A",
                                    rucioUrl=self.rucioUrl, rucioToken=self.rucioToken, scope=self.rucioScope)
        self.assertEqual(resp, {})

        with self.assertRaises(RuntimeError):
            listReplicationRules(["blah"], self.rucioAccount, grouping="Not valid!!!",
                                 rucioUrl=self.rucioUrl, rucioToken=self.rucioToken, scope=self.rucioScope)

        containers = [PU_CONT, CONT2, self.badDID]
        resp = listReplicationRules(containers, self.rucioAccount, grouping="A",
                                    rucioUrl=self.rucioUrl, rucioToken=self.rucioToken, scope=self.rucioScope)
        self.assertTrue(len(resp) == 3)
        self.assertTrue(PU_CONT in resp)
        self.assertTrue(isinstance(resp[CONT2], list))
        self.assertEqual(resp[self.badDID], [])

    def testGetBlocksAndSizeRucio(self):
        """
        Test the getBlocksAndSizeRucio function, which fetches all the blocks
        (in a container) and their sizes.
        """
        self.rucioToken, self.tokenValidity = getRucioToken(self.rucioAuthUrl, self.rucioAccount)
        # Test 1: no DIDs provided as input
        resp = getBlocksAndSizeRucio([], self.rucioUrl, self.rucioToken, self.rucioScope)
        self.assertEqual(resp, {})
        # Test 2: multiple valid/invalid DIDs provided as input
        containers = [PU_CONT, CONT3, self.badDID]
        resp = getBlocksAndSizeRucio(containers, self.rucioUrl, self.rucioToken, self.rucioScope)
        self.assertTrue(len(resp) == 3)
        self.assertTrue(resp[PU_CONT][PU_CONT_BLK]['blockSize'] > 0)
        self.assertTrue(isinstance(resp[PU_CONT][PU_CONT_BLK]['locations'], list))
        self.assertTrue(len(resp[CONT3]) > 0)
        self.assertIsNone(resp[self.badDID])

    def testStringDateToEpoch(self):
        """
        Test the stringDateToEpoch function
        """
        dateString = 'Thu, 29 Apr 2021 13:15:42 UTC'
        self.assertEqual(1619694942, stringDateToEpoch(dateString))
        dateString = 'Thu, 1 Jan 1970 01:00:00 UTC'
        self.assertEqual(0, stringDateToEpoch(dateString))


if __name__ == '__main__':
    unittest.main()



"""
_IsAllWorkflowsCompleted_

Oracle implementation of DBS3Buffer.IsAllWorkflowsCompleted
"""

from __future__ import print_function, division

from WMComponent.DBS3Buffer.MySQL.IsAllWorkflowsCompleted \
    import IsAllWorkflowsCompleted as MySQLIsAllWorkflowsCompleted

class IsAllWorkflowsCompleted(MySQLIsAllWorkflowsCompleted):
    """
    Oracle version
    """
    pass



#!/usr/bin/env python
"""
_LoadFromFilesetWorkflow_

Oracle implementation of Subscription.LoadFromFilesetWorkflow
"""

__all__ = []



from WMCore.WMBS.MySQL.Subscriptions.LoadFromFilesetWorkflow \
     import LoadFromFilesetWorkflow as LoadFromFilesetWorkflowMySQL

class LoadFromFilesetWorkflow(LoadFromFilesetWorkflowMySQL):
    pass



"""
_RunJobByStatus_

Monitoring DAO classes for Jobs in BossAir database.
It groups jobs in each sched_status and bossAir status and guarantee
all sched_status are always present in the output.
"""
from __future__ import print_function, division

from WMCore.Database.DBFormatter import DBFormatter


class RunJobByStatus(DBFormatter):
    sql = """
          SELECT bl_status.name AS sched_status, count(bl_runjob.sched_status) AS count, bl_runjob.status
            FROM bl_status
            LEFT OUTER JOIN bl_runjob ON bl_runjob.sched_status = bl_status.id
            GROUP BY bl_status.name, bl_runjob.status
          """

    def formatDict(self, results):
        """
        _formatDict_

        Creates a dictionary of active (status=1) and completed (status=0)
        jobs in BossAir with their sched_status and the amount of jobs in that status
        """
        formattedResults = DBFormatter.formatDict(self, results)

        results = {'active': {}, 'completed': {}}
        for res in formattedResults:
            results['active'].setdefault(res['sched_status'], 0)
            results['completed'].setdefault(res['sched_status'], 0)
            if res['status'] is None:
                pass  # job count is always 0 for this case
            elif int(res['status']) == 0:
                results['completed'][res['sched_status']] += int(res['count'])
            else:  # status = 1
                results['active'][res['sched_status']] += int(res['count'])

        return results

    def execute(self, conn=None, transaction=False):
        result = self.dbi.processData(self.sql, conn=conn, transaction=transaction)

        return self.formatDict(result)



#!/usr/bin/env python
"""
_ListSubsAndFilesetsFromWorkflow_

Creates a mapping between subscriptions id, fileset names and workflow
tasks.
"""
from __future__ import division

from WMCore.Database.DBFormatter import DBFormatter


class ListSubsAndFilesetsFromWorkflow(DBFormatter):
    """ Returns a list of tuples """
    sql = """SELECT wmbs_subscription.id as subID, wmbs_fileset.name AS fileset,
                 wmbs_workflow.task AS wfTask, wmbs_subscription.split_algo,
                 wmbs_sub_types.name AS subType
               FROM wmbs_subscription
                 INNER JOIN wmbs_fileset ON wmbs_subscription.fileset = wmbs_fileset.id
                 INNER JOIN wmbs_workflow ON wmbs_subscription.workflow = wmbs_workflow.id
                 INNER JOIN wmbs_sub_types ON wmbs_subscription.subtype = wmbs_sub_types.id
               WHERE wmbs_workflow.name = :workflow
               ORDER BY wmbs_fileset.name, wmbs_workflow.task"""

    def format(self, results):
        "Build a list of tuples"
        result = []
        results = DBFormatter.format(self, results)
        for item in results:
            result.append(tuple(item))

        return result

    def execute(self, workflow, returnTuple=False, conn=None, transaction=False):
        result = self.dbi.processData(self.sql, {"workflow": workflow},
                                      conn=conn, transaction=transaction)
        if returnTuple:
            # easier for comparison
            return self.format(result)
        else:
            return self.formatDict(result)



#!/usr/bin/env python
import os
import unittest
import shutil
from WMCore.Services.RequestDB.RequestDBReader import RequestDBReader
from WMCore.Services.RequestDB.RequestDBWriter import RequestDBWriter
from WMQuality.TestInitCouchApp import TestInitCouchApp
from WMCore_t.Services_t.WMStats_t.WMStatsDocGenerator import generate_reqmgr_schema


class T0RequestDBTest(unittest.TestCase):
    """
    """

    def setUp(self):
        """
        _setUp_
        """
        self.schema = []
        self.couchApps = ["T0Request"]
        self.testInit = TestInitCouchApp('RequestDBServiceTest')
        self.testInit.setLogging()
        self.testInit.setDatabaseConnection()
        self.testInit.setSchema(customModules=self.schema,
                                useDefault=False)
        dbName = 't0_requsetdb_t'
        self.testInit.setupCouch(dbName, *self.couchApps)
        reqDBURL = "%s/%s" % (self.testInit.couchUrl, dbName)
        self.requestWriter = RequestDBWriter(reqDBURL, self.couchApps[0])
        self.requestReader = RequestDBReader(reqDBURL, self.couchApps[0])
        self.requestWriter.defaultStale = {}
        self.requestReader.defaultStale = {}
        return

    def tearDown(self):
        """
        _tearDown_

        Drop all the WMBS tables.
        """
        self.testInit.tearDownCouch()

    def testRequestDBWriter(self):
        # test getWork
        schema = generate_reqmgr_schema()
        result = self.requestWriter.insertGenericRequest(schema[0])

        self.assertEqual(len(result), 1, 'insert fail');

        result = self.requestWriter.updateRequestStatus(schema[0]['RequestName'], "assigned")

        self.assertEqual(result, "Not allowed state: 'assigned'", 'update fail')
        self.assertEqual(self.requestWriter.updateRequestStatus("not_exist_schema", "new"),
                         'Error: document not found')

        allowedStates = ["Closed", "Merge", "AlcaSkim", "Harvesting",
                         "Processing Done", "completed"]
        for state in allowedStates:
            self.assertEqual(self.requestWriter.updateRequestStatus(schema[0]['RequestName'], state),
                             'OK')

        self.assertEqual(self.requestWriter.updateRequestStatus(schema[0]['RequestName'], "Processing Done"),
                         "Not allowed transition, from: 'completed' to 'Processing Done'")

        self.assertEqual(self.requestWriter.updateRequestStatus(schema[0]['RequestName'], "normal-archived"),
                         'OK')
        result = self.requestWriter.getRequestByStatus(["normal-archived"], False, 1)
        self.assertEqual(len(result), 1, "should be 1 but %s" % result)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
"""
_DeleteFiles_

Template for a DeleteFiles Step

Mostly borrowed from StageOut since they share a similar function

"""




from WMCore.WMSpec.Steps.Template import Template
from WMCore.WMSpec.Steps.Template import CoreHelper
from WMCore.WMSpec.ConfigSectionTree import nodeName



class DeleteFilesStepHelper(CoreHelper):
    """
    _StageOutStepHelper_

    Add API calls and helper methods to the basic WMStepHelper to specialise
    for StageOut tasks

    This is very similar to StageOut since they have essentially the same function

    """

    def disableRetries(self):
        """
            handy for testing, without the 10 minute retry loop
        """
        self.data.retryCount = 1
        self.data.retryDelay = 0

    def addOutputDestination(self, lfn):
        """
        Adds an out location to put a tarball of all logs
        """


class DeleteFiles(Template):
    """
    _DeleteFiles_

    Tools for creating a template DeleteFiles Step

    """

    def install(self, step):
        stepname = nodeName(step)
        step.stepType = "DeleteFiles"
        step.section_("logs")
        step.logcount = 0
        step.retryCount = 3
        step.retryDelay = 300


    def helper(self, step):
        """
        _helper_

        Wrap the WMStep provided in the CMSSW helper class that
        includes the ability to add and manipulate the details
        of a CMSSW workflow step

        """
        return DeleteFilesStepHelper(step)



#!/usr/bin/env python
"""
_ListThresholdsForSubmit_

Query WMBS and ResourceControl to determine how many jobs are still running so
that we can schedule jobs that have just been created.
"""

from future.utils import viewitems

from WMCore.Database.DBFormatter import DBFormatter


class ListThresholdsForSubmit(DBFormatter):
    sql = """SELECT wmbs_location.site_name AS site_name,
                    wmbs_location.pending_slots,
                    wmbs_location.running_slots,
                    wmbs_location.cms_name AS cms_name,
                    rc_threshold.max_slots,
                    rc_threshold.pending_slots AS task_pending_slots,
                    wmbs_sub_types.name AS task_type,
                    job_count.job_status,
                    job_count.jobs,
                    job_count.wf_highest_priority,
                    wmbs_sub_types.priority,
                    wmbs_location_state.name AS state,
                    wmbs_location.plugin AS plugin,
                    wmbs_location.state_time AS state_time
                    FROM wmbs_location
               INNER JOIN rc_threshold ON
                 wmbs_location.id = rc_threshold.site_id
               INNER JOIN wmbs_sub_types ON
                 rc_threshold.sub_type_id = wmbs_sub_types.id
               INNER JOIN wmbs_location_state ON
                wmbs_location_state.id = wmbs_location.state
               LEFT OUTER JOIN
                 (SELECT wmbs_job.location AS location,
                         wmbs_subscription.subtype AS subtype,
                         COUNT(wmbs_job.id) as jobs,
                         MAX(wmbs_workflow.priority) as wf_highest_priority,
                         bl_status.name as job_status
                         FROM wmbs_job
                    INNER JOIN wmbs_jobgroup ON
                      wmbs_job.jobgroup = wmbs_jobgroup.id
                    INNER JOIN wmbs_subscription ON
                      wmbs_jobgroup.subscription = wmbs_subscription.id
                    INNER JOIN wmbs_job_state ON
                      wmbs_job.state = wmbs_job_state.id
                    INNER JOIN bl_runjob ON
                      bl_runjob.wmbs_id = wmbs_job.id
                    INNER JOIN bl_status ON
                      bl_status.id = bl_runjob.sched_status
                    INNER JOIN wmbs_workflow ON
                      wmbs_workflow.id = wmbs_subscription.workflow
                  WHERE wmbs_job_state.name = 'executing' AND
                        bl_runjob.status = '1'
                  GROUP BY wmbs_job.location, wmbs_subscription.subtype,
                           bl_status.name) job_count ON
                  wmbs_location.id = job_count.location AND
                  wmbs_sub_types.id = job_count.subtype
               ORDER BY wmbs_sub_types.priority DESC"""

    seSql = """SELECT wl.site_name AS site_name,
                      wpnn.pnn AS pnn
               FROM wmbs_location wl
                      INNER JOIN wmbs_location_pnns wls ON
                          wls.location = wl.id
                      INNER JOIN wmbs_pnns wpnn ON
                          wpnn.id = wls.pnn
            """

    def format(self, results, storageElements):
        """
        _format_

        Add up totals from various rows and combine that together into a single
        data structure.
        """
        results = DBFormatter.formatDict(self, results)
        storageElements = DBFormatter.formatDict(self, storageElements)

        mappedPNNs = {}
        for pnn in storageElements:
            if pnn['site_name'] not in mappedPNNs:
                mappedPNNs[pnn['site_name']] = []
            mappedPNNs[pnn['site_name']].append(pnn['pnn'])

        formattedResults = {}
        for result in results:
            siteName = result['site_name']
            taskType = result['task_type']

            task_pending_jobs = 0
            task_running_jobs = 0
            if result['job_status']:
                module = __import__("WMCore.BossAir.Plugins.%s" % result['plugin'],
                                    globals(), locals(), [result['plugin']])
                plugIn = getattr(module, result['plugin'])
                status = plugIn.stateMap().get(result['job_status'])
                if status == 'Pending':
                    task_pending_jobs += result['jobs']
                elif status == 'Running':
                    task_running_jobs += result['jobs']

            if siteName not in formattedResults:
                siteInfo = {}
                siteInfo['pnns'] = mappedPNNs.get(siteName, [])
                siteInfo['state'] = result['state']
                siteInfo['state_time'] = result['state_time']
                siteInfo['total_pending_slots'] = result['pending_slots']
                siteInfo['total_running_slots'] = result['running_slots']
                siteInfo['total_pending_jobs'] = task_pending_jobs
                siteInfo['total_running_jobs'] = task_running_jobs
                siteInfo['thresholds'] = {}
                formattedResults[siteName] = siteInfo
            else:
                formattedResults[siteName]['total_pending_jobs'] += task_pending_jobs
                formattedResults[siteName]['total_running_jobs'] += task_running_jobs

            if taskType not in formattedResults[siteName]['thresholds']:
                threshold = {}
                threshold[taskType] = {}
                threshold[taskType]['max_slots'] = result['max_slots']
                threshold[taskType]['pending_slots'] = result['task_pending_slots']
                threshold[taskType]['priority'] = result['priority']
                threshold[taskType]['wf_highest_priority'] = result['wf_highest_priority']
                threshold[taskType]['task_running_jobs'] = task_running_jobs
                threshold[taskType]['task_pending_jobs'] = task_pending_jobs
                formattedResults[siteName]['thresholds'].update(threshold)
            else:
                formattedResults[siteName]['thresholds'][taskType]['task_pending_jobs'] += task_pending_jobs
                formattedResults[siteName]['thresholds'][taskType]['task_running_jobs'] += task_running_jobs

        return formattedResults

    def formatTable(self, formattedResults):
        """
        _format_

        Combine together the total we received from the assigned and unassigned
        queries into a single datastructure.
        """
        results = []
        for k, v in viewitems(formattedResults):
            item = {}
            item['site'] = k
            item['data'] = []
            for ck, cv in viewitems(v):
                childItem = {}
                childItem['type'] = ck
                childItem.update(cv)
                item['data'].append(childItem)
            results.append(item)
        return {'results': results}

    def execute(self, conn=None, transaction=False, tableFormat=False):
        results = self.dbi.processData(self.sql, conn=conn, transaction=transaction)
        pnns = self.dbi.processData(self.seSql, conn=conn, transaction=transaction)
        results = self.format(results, pnns)

        if tableFormat:
            return self.formatTable(results)
        else:
            return results



from builtins import object
NOT_EXIST_DATASET = 'thisdoesntexist'
PILEUP_DATASET = '/HighPileUp/Run2011A-v1/RAW'

SITES = ['T2_XX_SiteA', 'T2_XX_SiteB', 'T2_XX_SiteC']

global _BLOCK_LOCATIONS

_BLOCK_LOCATIONS = {}

def getSites(block):
    if block in _BLOCK_LOCATIONS:
        return _BLOCK_LOCATIONS[block]

    if block.split('#')[0] == PILEUP_DATASET:
        # Pileup is at a single site
        sites = ['T2_XX_SiteC']
        _BLOCK_LOCATIONS[block] = sites
    elif block.endswith('#1'):
        sites  = ['T2_XX_SiteA']
        _BLOCK_LOCATIONS[block] = sites
    elif block.endswith('#2'):
        sites = ['T2_XX_SiteA', 'T2_XX_SiteB']
        _BLOCK_LOCATIONS[block] = sites
    else:
        sites = ['T2_XX_SiteA', 'T2_XX_SiteB','T2_XX_SiteC']
    return sites

# This can be replace to PhEDEx subscription call
# emulator
def moveBlock(blockSites):
    _BLOCK_LOCATIONS.update(blockSites)

class GlobalParams(object):

    # WARNING: do not change the default value:
    # unittests will break.
    # If they are needed to be changed in some unittests
    # or other situation, use the setter.
    # If they are changed in the unittest, reset to default values in tearDown
    # (otherwise buildbot test will have unexpected result,
    # since it will run several unittests in one process)

    _num_of_blocks_per_dataset = 2
    _num_of_files_per_block = 5
    _num_of_runs_per_file = 1
    _num_of_lumis_per_block = 2
    _num_of_events_per_file = 1000
    _size_of_file = 20000000
    _blocks_open_for_writing = False

    @staticmethod
    def numOfBlocksPerDataset():
        return GlobalParams._num_of_blocks_per_dataset

    @staticmethod
    def numOfFilesPerBlock():
        return GlobalParams._num_of_files_per_block

    @staticmethod
    def numOfRunsPerFile():
        return GlobalParams._num_of_runs_per_file

    @staticmethod
    def numOfEventsPerFile():
        return GlobalParams._num_of_events_per_file

    @staticmethod
    def numOfLumisPerBlock():
        #It's really the number of lumis per file
        return GlobalParams._num_of_lumis_per_block

    @staticmethod
    def sizeOfFile():
        return GlobalParams._size_of_file

    @staticmethod
    def setNumOfBlocksPerDataset(numOfBlocksPerDataset):
        GlobalParams._num_of_blocks_per_dataset = numOfBlocksPerDataset

    @staticmethod
    def setNumOfFilesPerBlock(numOfFilesPerBlock):
        GlobalParams._num_of_files_per_block = numOfFilesPerBlock

    @staticmethod
    def setNumOfRunsPerFile(numOfRunsPerFile):
        GlobalParams._num_of_runs_per_file = numOfRunsPerFile

    @staticmethod
    def setNumOfEventsPerFile(numOfEventsPerFile):
        GlobalParams._num_of_events_per_file = numOfEventsPerFile

    @staticmethod
    def setNumOfLumisPerBlock(numOfLumisPerBlock):
        GlobalParams._num_of_lumis_per_block = numOfLumisPerBlock

    @staticmethod
    def setSizeOfFile(sizeOfFile):
        GlobalParams._size_of_file = sizeOfFile

    @staticmethod
    def getRunNumberForBlock(blockName):
        #assumes blockName is contains number after '#'
        return int(blockName.split('#')[-1])

    @staticmethod
    def resetParams():
        """
        reset the parameter to default value
        """
        GlobalParams._num_of_blocks_per_dataset = 2
        GlobalParams._num_of_files_per_block = 5
        GlobalParams._num_of_runs_per_file = 1
        GlobalParams._num_of_lumis_per_block = 2
        GlobalParams._num_of_events_per_file = 1000
        GlobalParams._size_of_file = 20000000
        GlobalParams._blocks_open_for_writing = False



from __future__ import print_function

import logging
import threading
import traceback
import unittest

import cherrypy
import cherrypy.process.wspbus as cherrybus

# decorator import for RESTServer setup
from WMCore.WebTools.Root import Root


class RESTBaseUnitTest(unittest.TestCase):
    def setUp(self, initRoot=True):
        # default set
        self.schemaModules = []

        self.initialize()
        if self.schemaModules:
            import warnings
            warnings.warn("use RESTAndCouchUnitTest instead", DeprecationWarning)
            from WMQuality.TestInitCouchApp import TestInitCouchApp
            self.testInit = TestInitCouchApp(__file__)
            self.testInit.setLogging()  # logLevel = logging.SQLDEBUG
            self.testInit.setDatabaseConnection(destroyAllDatabase=True)
            self.testInit.setSchema(customModules=self.schemaModules,
                                    useDefault=False)
            # Now pull the dbURL from the factory
            # I prefer this method because the factory has better error handling
            # Also because then you know everything is the same
            myThread = threading.currentThread()
            self.config.setDBUrl(myThread.dbFactory.dburl)

        logging.info("This is our config: %s" % self.config)

        self.initRoot = initRoot
        if initRoot:
            self.rt = Root(self.config, testName=self._testMethodName)
            try:
                self.rt.start(blocking=False)
            except RuntimeError as e:
                # there appears to be worker threads from a previous test
                # hanging out. Try to slay them so that we can keep going
                print("Failed to load cherrypy with exception: %s\n" % e)
                print("The threads are: \n%s\n" % threading.enumerate())
                print("The previous test was %s\n" % self.rt.getLastTest())
                print(traceback.format_exc())
                self.rt.stop()
                raise e

        return

    def tearDown(self):
        if self.initRoot:
            self.rt.stop()
            self.rt.setLastTest()
            # there was a ton of racy failures in REST tools because of
            # how much global state cherrypy has. this resets it

            # Also, it sucks I had to copy/paste this from
            # https://bitbucket.org/cherrypy/cherrypy/src/9720342ad159/cherrypy/__init__.py
            # but reload() doesn't have the right semantics

            cherrybus.bus = cherrybus.Bus()
            cherrypy.engine = cherrybus.bus
            # This class has apparently been deprecated in the newer versions
            # cherrypy.engine.timeout_monitor = cherrypy._TimeoutMonitor(cherrypy.engine)
            # cherrypy.engine.timeout_monitor.subscribe()

            cherrypy.engine.autoreload = cherrypy.process.plugins.Autoreloader(cherrypy.engine)
            cherrypy.engine.autoreload.subscribe()

            cherrypy.engine.thread_manager = cherrypy.process.plugins.ThreadManager(cherrypy.engine)
            cherrypy.engine.thread_manager.subscribe()

            cherrypy.engine.signal_handler = cherrypy.process.plugins.SignalHandler(cherrypy.engine)
            cherrypy.engine.subscribe('log', cherrypy._buslog)

            from cherrypy import _cpserver
            cherrypy.server = _cpserver.Server()
            cherrypy.server.subscribe()
            cherrypy.checker = cherrypy._cpchecker.Checker()
            cherrypy.engine.subscribe('start', cherrypy.checker)

        if self.schemaModules:
            self.testInit.clearDatabase()
        self.config = None
        return

    def initialize(self):
        """
        i.e.

        self.config = DefaultConfig('WMCore.WebTools.RESTModel')
        self.config.setDBUrl('sqlite://')
        self.schemaModules = ['WMCore.ThreadPool', 'WMCore.WMBS']
        """

        message = "initialize method has to be implemented, self.restModel, self.schemaModules needs to be set"
        raise NotImplementedError(message)



#!/usr/bin/env python
"""
__ThreadSlave__

Base class for implementing slave objects.
slave objects inherit from this and overload
the __call__ method. This method is (usually) passed
a list of parmeters used to complete a certain tasks.
parameters is an object that is pickled into the database
when it is not processed by a thread. It is thus possible
to pass arbitrary objects if needed through the parameters
attribute.
"""





from future import standard_library
standard_library.install_aliases()
from builtins import str, object

import base64
import logging
import threading
import os
import pickle

from WMCore.Database.Transaction import Transaction
from WMCore.WMFactory import WMFactory

from Utils.PythonVersion import PY3
from Utils.Utilities import encodeUnicodeToBytes

class ThreadSlave(object):
    """
    __ThreadSlave__

    Base class for implementing slave objects.
    slave objects inherit from this and overload
    the __call__ method. This method is (usually) passed
    a list of parmeters used to complete a certain tasks.
    parameters is an object that is pickled into the database
    when it is not processed by a thread. It is thus possible
    to pass arbitrary objects if needed through the parameters
    attribute.
    """

    def __init__(self ):
        """
        The constructor creates separate instances
        for objects that we do not want to share.
        Objects that are used as a thread slave
        Inherit from this to ensure they do not
        conflict with these methods.

        If needed the developer can add additional
        objects as it sees fit in the threadslave
        class that inherits from this.
        """
        self.args = {}
        # we also keep a reference to our component (we can
        # use this for read only things in the argument list).
        # assign this later
        self.component = None
        #a slave is created in its master thread so we can exploit
        #this to get a reference to its dbfactory object.
        myThread = threading.currentThread()
        self.dbFactory = myThread.dbFactory
        # get the procid from the mainthread msg service
        # if we use this in testing it might not be there.
        self.procid = 0

        # we can potentially use mapping from messages to handlers
        # to have one thread handling multiple message types.
        self.messages = {}
        # start loading some objects we need in this thread only
        factory = WMFactory("threadPool", "WMCore.ThreadPool."+ \
            myThread.dialect)
        self.query = factory.loadObject("Queries")

        #NOTE: this is not a new thread so we need to propagate
        #NOTE: the arguments we want to carry over to the current thread
        #NOTE: object using the initInThread method later on.

    def initInThread(self):
        """
        ThreadSlave (subclasses) are returned to a queue when no work
        is available and can be activated by a new thread when work
        is entered in the queue. This method is called when the ThreadSlave
        (subclass) is associated to a new thread.
        """

        # we need to call this method only when it is called within a thread
        # otherwise these parameters are not accissible in the thread used
        # to call this threadslave.

        myThread = threading.currentThread()
        # we are now in our own thread, so we can pass the dbFactory reference
        # to here:
        myThread.dbFactory = self.dbFactory

        if self.component.config.CoreDatabase.dialect:
            myThread.dialect = self.component.config.CoreDatabase.dialect
        else:
            # FIXME we aren't using the DIALECT environment variable anymore
            myThread.dialect = os.getenv("DIALECT")

        #TODO: remove as much as possible logging statements or make them debug
        myThread.logger = logging.getLogger()

        logging.info("THREAD: Initializing default database")
        logging.info("THREAD: Check if connection is through socket")
        logging.info("THREAD: Building database connection string")
        # we ensured that we use the dbFactory object from our parent
        # thread so we have only one engine in the application.
        myThread.dbi = myThread.dbFactory.connect()
        logging.info("THREAD: Initializing default transaction")
        myThread.transaction = Transaction(myThread.dbi)
        logging.info("THREAD: Loading backend")
        logging.info("THREAD constructor finished")

    # FIXME: enable the (configurable) possibility to retrieve
    # work in bulk (handy when we need to update databases)
    def retrieveWork(self):
        """
        _retrieveWork_

        If activated this threadlsave retrieves work. It retrieves
        it from the persistent thread pool and changes the state
        from queued to process.
        """
        myThread = threading.currentThread()
        # we only want to intitiate thread related issues once per thread.
        # this checks if our thread has the dbi attributes.
        if not hasattr(myThread,"dbi"):
            self.initInThread()
        else:
            # init creates a transaction that will call begin.
            myThread.transaction.begin()
        args = {'thread_pool_id' : self.args['thread_pool_id'], \
            'component' : self.args['componentName']}
        result = self.query.selectWork(args, \
            self.args['thread_pool_table_buffer_out'])
        # we might need to look into multiple buffers and move work to find it.
        # from keeping track of the number of messages for us we know it
        # is there.
        if result[0] == None:
            self.query.moveWorkToBufferOut(args, \
                self.args['thread_pool_table'], \
                self.args['thread_pool_table_buffer_out'], \
                self.args['thread_pool_buffer_size'])
        result = self.query.selectWork(args, \
            self.args['thread_pool_table_buffer_out'])
        if result[0] == None :
            self.query.moveWorkToBufferOut(args, \
                self.args['thread_pool_table_buffer_in'], \
                self.args['thread_pool_table_buffer_out'], \
                self.args['thread_pool_buffer_size'])
        result = self.query.selectWork(args, \
                self.args['thread_pool_table_buffer_out'])

        if result[0] == None:
            # FIXME: make proper exception
            raise Exception("ERROR: How can that be!!")
        logging.debug("THREAD: Retrieved Work with id: "+str(result[0]) )
        myThread.workId = str(result[0])
        # get the actual work now:
        result = self.query.retrieveWork({'id':myThread.workId}, \
            self.args['thread_pool_table_buffer_out'])
        self.query.tagWork({'id' : myThread.workId}, \
            self.args['thread_pool_table_buffer_out'])
        # we commit here because if the component crashes this is where
        # if will look for lost threads (the ones that are in the process state
        myThread.transaction.commit()
        base64_decoder = base64.decodebytes if PY3 else base64.decodestring
        return (result[1], pickle.loads(base64_decoder(encodeUnicodeToBytes(result[2]))))

    def removeWork(self):
        """
        _removeWork_

        Once the work is finished the entry is removed from the queue.
        """
        myThread = threading.currentThread()
        myThread.transaction.begin()
        self.query.removeWork({'id' : myThread.workId}, \
            self.args['thread_pool_table_buffer_out'])
        # this method is called once the thread is finished and
        # we commit everything.
        myThread.transaction.commit()

        logging.debug("THREAD: Removed Work")

    def __call__(self, parameters):
        """
        _call_

        components using a (or multiple) threadpools overload
        this method. They inherit from this class and make sure
        its constructor is called, and overload this call method.
        """
        logging.error("I am a placeholder please overload me to handle parameters : %s", str(parameters))



"""
Unit tests for MicroService.py module

Author: Valentin Kuznetsov <vkuznet [AT] gmail [DOT] com>
"""

from __future__ import division, print_function

import unittest
import cherrypy
import gzip
import json
from WMCore_t.MicroService_t import TestConfig
from WMCore.MicroService.Service.RestApiHub import RestApiHub
from WMCore.MicroService.Tools.Common import cert, ckey
from WMCore.Services.pycurl_manager import RequestHandler
from Utils.Utilities import decodeBytesToUnicode


def gzipDecompress(payload):
    """Util to Gzip decompress a given data object"""
    if isinstance(payload, bytes):
        payload = gzip.decompress(payload)
        payload = decodeBytesToUnicode(payload)
        return json.loads(payload)
    return payload


class ServiceManager(object):
    """
    Initialize ServiceManager class
    """

    def __init__(self, config=None):
        self.config = config
        self.appname = 'test'  # keep it since it is used by XMLFormat(self.app.appname))

    def status(self, serviceName=None, **kwargs):
        "Return current status about our service"
        print("### CALL status API with service name %s" % serviceName)
        data = {'status': "OK", "api": "status"}
        if kwargs:
            data.update(kwargs)
        return data

    def info(self, reqName, **kwargs):
        "Return current status about our service"
        print("### CALL info API with request name %s" % reqName)
        data = {'status': "OK", "api": "info"}
        if kwargs:
            data.update(kwargs)
        return data


class MicroServiceTest(unittest.TestCase):
    "Unit test for MicroService module"

    def setUp(self):
        "Setup MicroService for testing"
        self.managerName = "ServiceManager"
        config = TestConfig
        manager = 'WMCore_t.MicroService_t.MicroService_t.%s' % self.managerName
        config.views.data.manager = manager
        config.manager = manager
        mount = '/microservice/data'
        self.mgr = RequestHandler()
        self.port = config.main.port
        self.url = 'http://localhost:%s%s' % (self.port, mount)
        cherrypy.config["server.socket_port"] = self.port
        self.app = ServiceManager(config)
        self.server = RestApiHub(self.app, config, mount)
        cherrypy.tree.mount(self.server, mount)
        cherrypy.engine.start()
        # implicitly request data compressed with gzip (default in RequestHandler class)
        self.noEncHeader = {'Accept': 'application/json'}
        # explicitly request data uncompressed
        self.identityEncHeader = {'Accept': 'application/json', 'Accept-Encoding': 'identity'}
        # explicitly request data compressed with gzip
        self.gzipEncHeader = {'Accept': 'application/json', 'Accept-Encoding': 'gzip'}

    def tearDown(self):
        "Tear down MicroService"
        cherrypy.engine.stop()
        cherrypy.engine.exit()

    def testGetStatus(self):
        "Test function for getting state of the MicroService"
        api = "status"
        url = '%s/%s' % (self.url, api)
        params = {}
        data = self.mgr.getdata(url, params=params, headers=self.noEncHeader, encode=True, decode=True)
        data = gzipDecompress(data)
        self.assertEqual(data['result'][0]['microservice'], self.managerName)
        self.assertEqual(data['result'][0]['api'], api)

        params = {"service": "transferor"}
        data = self.mgr.getdata(url, params=params, headers=self.noEncHeader, encode=True, decode=True)
        data = gzipDecompress(data)
        self.assertEqual(data['result'][0]['microservice'], self.managerName)
        self.assertEqual(data['result'][0]['api'], api)

    def testGetStatusIdentity(self):
        "Test function for getting state of the MicroService"
        api = "status"
        url = '%s/%s' % (self.url, api)
        params = {}
        data = self.mgr.getdata(url, params=params, headers=self.identityEncHeader, encode=True, decode=True)
        self.assertEqual(data['result'][0]['microservice'], self.managerName)
        self.assertEqual(data['result'][0]['api'], api)

        params = {"service": "transferor"}
        data = self.mgr.getdata(url, params=params, headers=self.identityEncHeader, encode=True, decode=True)
        self.assertEqual(data['result'][0]['microservice'], self.managerName)
        self.assertEqual(data['result'][0]['api'], api)

    def testGetInfo(self):
        "Test function for getting state of the MicroService"
        api = "info"
        url = '%s/%s' % (self.url, api)
        params = {}
        data = self.mgr.getdata(url, params=params, encode=True, decode=True)
        data = gzipDecompress(data)
        self.assertEqual(data['result'][0]['microservice'], self.managerName)
        self.assertEqual(data['result'][0]['api'], api)

        params = {"request": "fake_request_name"}
        data = self.mgr.getdata(url, params=params, encode=True, decode=True)
        data = gzipDecompress(data)
        self.assertEqual(data['result'][0]['microservice'], self.managerName)
        self.assertEqual(data['result'][0]['api'], api)

    def testGetInfoGZipped(self):
        "Test function for getting state of the MicroService"
        api = "status"
        url = '%s/%s' % (self.url, api)
        params = {}
        # headers = {'Content-Type': 'application/json', 'Accept-Encoding': 'gzip'}
        data = self.mgr.getdata(url, params=params, headers=self.gzipEncHeader, encode=True, decode=True)
        # data = self.mgr.getdata(url, params=params, encode=True, decode=False)
        # data = gzipDecompress(data)
        self.assertEqual(data['result'][0]['microservice'], self.managerName)
        self.assertEqual(data['result'][0]['api'], api)

        params = {"request": "fake_request_name"}
        data = self.mgr.getdata(url, params=params, headers=self.gzipEncHeader, encode=True, decode=True)
        # data = gzipDecompress(data)
        self.assertEqual(data['result'][0]['microservice'], self.managerName)
        self.assertEqual(data['result'][0]['api'], api)

    def testPostCall(self):
        "Test function for getting state of the MicroService"
        api = "status"
        url = self.url + "/%s" % api
        params = {"request": "fake_request_name"}
        headers = {'Content-Type': 'application/json'}
        data = self.mgr.getdata(url, params=params, headers=headers, verb='POST',
                                cert=cert(), ckey=ckey(), encode=True, decode=True)
        self.assertDictEqual(data['result'][0], {'status': 'OK', 'api': 'info'})

    def testPostCallGZipped(self):
        "Test function for getting state of the MicroService"
        api = "status"
        url = self.url + "/%s" % api
        params = {"request": "fake_request_name"}
        headers = {'Content-Type': 'application/json', 'Accept-Encoding': 'gzip'}
        data = self.mgr.getdata(url, params=params, headers=headers, verb='POST',
                                cert=cert(), ckey=ckey(), encode=True, decode=True)
        self.assertDictEqual(data['result'][0], {'status': 'OK', 'api': 'info'})


if __name__ == '__main__':
    unittest.main()
